//----------------------------------------------------------------------------
// 
//	file: netlayer.cc
// 
//	function: network layer routines : implements a single neural 
//		network layer 
//	author:   Yoonsuck Choe <yschoe@cs.utexas.edu>
//	creation: Wed Jul 22 04:37:01 CDT 1998
// 
//	$Revision: 1.16.1.16 $ $Date: 2001/11/30 20:24:15 $
//	$Log: netlayer.cc,v $
//	Revision 1.16.1.16  2001/11/30 20:24:15  yschoe
//	pglissom v1.0 release
//
//	Revision 1.16.1.15  2000/02/22 03:52:36  yschoe
//	1. when dumping weight, save only in the range where there are actual
//	   connections.
//	2.
//	   save C.O.G. added
//
//	Revision 1.16.1.14  2000/02/13 10:29:43  yschoe
//	*** empty log message ***
//
//	Revision 1.16.1.13  2000/01/30 14:56:21  yschoe
//	*** empty log message ***
//
//	Revision 1.16.1.12  1999/12/22 05:29:48  yschoe
//	or pref stuff, kill overlapping lat, etc.
//
//	Revision 1.16.1.11  1999/02/09 07:42:25  yschoe
//	1. circular aff weights (i.e. RFs)
//	2. instar/outstar option for exc weight :get_lat_excw
//	3. or_pref calculation fixed and util routines for or_pref added
//	4. dumping aff/lat weights routine added.
//
//	Revision 1.16.1.10  1999/01/10 04:16:45  yschoe
//	1. added or_pref calculation and print routines
//	2. added center of gravity calculation
//
//	Revision 1.16.1.9  1998/11/29 08:53:58  yschoe
//	fixed get_wgt routines -- base substraction was incorrect
//	, and added display option for instar or outstar normlaization in the
//	get inh weight function. The end result seems comparable.
//
//	Revision 1.16.1.8  1998/11/27 06:03:38  yschoe
//	1. check for killed weights.
//	2. added kill_overlapping_weights()
//	3. normalization by sum of wgt (not sqrt of sum of sqrs)
//		-- the inline function was not being overridden in
//		   derived classes!!
//
//	Revision 1.16.1.7  1998/11/17 12:40:42  yschoe
//	cleaned up weight normalization and added normalize_instar_weights().
//
//	Revision 1.16.1.5  1998/10/26 03:48:36  yschoe
//	1. moved reset_matrix() call to immediately necessary locations
//	2. resize_lat_exc_rad() : only allow reducing it.
//			fixed bug when calculating offset for weight copy
//	3. added command line argument style calls to kill_lat_weights() and
//	   resize_lat_exc_rad() : for use with simenv class
//	4. same with update_sig()
//	5. fixed setup_weight() to handle multiple input layers
//	6. get_xxx_weights() added for copying weights for a given unit (i,j) for
//	   use with display routine from other classes
//
//	Revision 1.16.1.4  1998/10/16 10:08:04  yschoe
//	1. added randomness parameter for weight setup.
//	2. reorganized activate() and introduced calc_contrib_sum()
//	3. separated set_parameter() parsing routine and put it in blackboard class.
//
//	Revision 1.16.1.3  1998/10/05 19:48:46  yschoe
//	1. added get_inh|exc|affw_matrix() functions
//	2. added setup_weight() functions
//	3. separated calc_affresp and calc_latresp from calc_netinput
//	4. set_matrix() for gaussian weight setup
//
//	Revision 1.16.1.2  1998/09/26 10:32:32  yschoe
//	1. fixed bug where sig_diff was not initialized, thus resulted in
//	   failure of first iteration.
//	2. now activate(1) only calculates afferent contrib, and
//	   activate(0) only calculates lateral resp.
//
//	Revision 1.16.1.1  1998/09/10 07:03:06  yschoe
//	removed a single '-' from the act_func() function
//
//	Revision 1.16  1998/09/06 05:59:02  yschoe
//	aff_contris is now fixed.
//
//	Revision 1.15  1998/09/04 08:45:08  yschoe
//	move redundant if from inside to outside of loop
//
//	Revision 1.14  1998/08/26 02:42:09  yschoe
//	Made inline functions non-virtual.
//
//	Revision 1.13  1998/08/22 08:18:50  yschoe
//	added net_aff/exc/inh_input and changed
//	interface to activate(). calc_netinput(), learn(), and normalize_weights()
//	were made outstar.
//
//	Revision 1.12  1998/08/14 17:44:29  yschoe
//	1. fixed bug in resize_lat_exc_rad().
//	2. added kill_lat_weights.
//
//	Revision 1.11  1998/08/14 06:50:29  yschoe
//	1. fixed little cosmetic bug in destructor
//	2. added resize_lat_exc_rad()
//
//	Revision 1.10  1998/08/11 17:01:09  yschoe
//	1. Modularized norm_sum() and norm_factor() for easier
//	extension of weight normalization.
//	2. Added reset_matrix() and reset_output()
//
//	Revision 1.9  1998/08/02 11:05:30  yschoe
//	1. calc_netinput() separated from activate().
//	2. fixed set_parameter(), get_parameter() and added different interfaces.
//
//	Revision 1.8  1998/07/30 00:18:53  yschoe
//	1. fixed destructor, init_weights(),
//	   normalize_weights() to use precalculated ranges.
//	2. fixed serious bug in activate(), weighted sum loop.
//	3. added learn(), delta(), and get_parameter().
//
//	Revision 1.7  1998/07/28 06:44:22  yschoe
//	1. init_weights() can now take a range [low,high)
//	2. activate() was added
//	3. set_parameter() was added --> will be removed later (maybe)
//
//	Revision 1.6  1998/07/26 23:58:50  yschoe
//	1. fixed bug : initializing aff_..[].
//	2. added update_blackboard() for setting up afferent weight entries
//	   when new input layer was added.
//	3. fixed net_bb->add (see 1.)
//	4. added normalize_weights().
//	5. separated calc_rf_ranges from add_inp_layer() so that
//	   it could be done before memory allocation (in alloc_first()).
//
//	Revision 1.5  1998/07/26 02:11:55  yschoe
//	init_weights(), act_func() added.
//	sig_low, sig_high added and added to blackboard.
//	must add normalization code.
//
//	Revision 1.4  1998/07/24 11:22:08  yschoe
//	save() load() and list() blackboard
//	routines added
//
//	Revision 1.3  1998/07/24 08:22:09  yschoe
//	fixed bug in add_inp_layer.
//	added add_lat_connections.
//	variables were added to blackboard(this was made a separate function)
//
//	Revision 1.2  1998/07/23 08:45:32  yschoe
//	alloc/dealloc of weights, setting up variables done. 
//	Todo: adding everything to blackboard, initializing weights, 
//	and all the real action stuff.
//
//	Revision 1.1  1998/07/23 01:07:05  yschoe
//	jul22 stable version (except netlayer)
//
//
//	header changed (Revision 1.14)
//----------------------------------------------------------------------------

#include "netlayer.h"
#include "parser.h"
#include "string.h"
#include "unistd.h"

//----------------------------------------------------------------------------
//	netlayer() 
//	  - Constructor
//	  - Only deal with stuff that is directly related to the 
//	  	network topology. Drop all stuff such as threshold, etc.
//----------------------------------------------------------------------------
netlayer::netlayer(char *name, int cols, int rows, 
		 int exc_rad, int inh_rad,
                 int max_inp_layers, netlayer *feeding,
		 int inp_cols, int inp_rows, int  aff_rad, int padded_flag) {

  // init some pseudo constants
  MIN_REAL = 1.0E-40;
  strncpy(netlayer::name,name,strlen(name)+1);

  // default values : for backward compatibility
  recursive_lateral_flag = 1;

  // network dimension
  netlayer::cols = cols; 
  netlayer::rows = rows; 

  // number of input layers and pointer to those areas.
  netlayer::max_inp_layers = max_inp_layers;
  netlayer::num_inp_layers = 0;

  // RF and lateral connection radius
  netlayer::lat_exc_rad = exc_rad;
  netlayer::lat_inh_rad = inh_rad;

  // padded_flag : is the retina padded? -- wasn't here. this was fixed: 00nov9
  netlayer::padded_flag =  padded_flag;

  // check for input/output dimension restrictions
  
  /*
  if (inp_cols!=0 && inp_rows!=0) {
     if (MAX(inp_cols,cols)%MIN(inp_cols,cols)!=0 ||
         MAX(inp_rows,rows)%MIN(inp_rows,rows)!=0 ) {
	error_exit("netlayer::netlayer","input area and output area size "
		"must be of integral proportion.");
     }
  }
  */

  // These are needed before anything is allocated 
  //	- taken out from alloc_first()
  netlayer::aff_rads 	= new int [max_inp_layers];
  netlayer::rf_ranges   = new NET_RF_Range** [max_inp_layers];


  // calculate afferent RF ranges first -- before allocating memory
  // 	- taken out from add_inp_layer()
  calc_rf_ranges(num_inp_layers,cols,rows,inp_cols,inp_rows,aff_rad);

  // allocate memory according to given network dimensions
  alloc_first(cols, rows, max_inp_layers, exc_rad, inh_rad); 

  // add connection to the given input layer : allocation is done within
  if (feeding != NULL) {
  	add_inp_layer(feeding,inp_cols,inp_rows,aff_rad);
  }

  // add all variables to the blackboard
  init_blackboard(name);

}

//----------------------------------------------------------------------------
//	~netlayer() 
//	  - descructor
//----------------------------------------------------------------------------
netlayer::~netlayer() {

  // dealloc stuff allocated by alloc_first()
  delete_matrix(cols,rows,net_input);
  delete_matrix(cols,rows,net_aff_input);
  delete_matrix(cols,rows,net_exc_input);
  delete_matrix(cols,rows,net_inh_input);
  delete_matrix(cols,rows,wgt_sum);
  delete_matrix(cols,rows,output);
  delete_matrix(cols,rows,prev_output);
  delete_matrix(cols,rows,or_pref);
  delete_matrix(cols,rows,or_selectivity);
  delete_matrix(MAX(cols,feedings[0]->cols),
		MAX(rows,feedings[0]->rows), ref_pattern);

  // dealloc afferent weights
  delete [] aff_rads;
  delete [] aff_contris;
  delete [] aff_noises;
  delete [] alpha_affs;
  delete [] feedings; // don't dealloc what's pointed by feedings[]
  for (int k=0; k<num_inp_layers; ++k) 
     for (int i=0; i<cols; ++i) 
	for (int j=0; j<rows; ++j) 
	   delete_matrix(rf_ranges[k][i][j].cols,
			 rf_ranges[k][i][j].rows,wgt[i][j].aff_ws[k]);
  // dealloc cog stuff
  for (int k=0; k<num_inp_layers; ++k) 
     for (int i=0; i<cols; ++i) {
	for (int j=0; j<rows; ++j) {
	   delete [] cogs[k][i];
	}	
	delete [] cogs[k];
     }

  delete [] cogs;

  // dealloc lateral weights
  for (int i=0; i<cols; ++i) 
     for (int j=0; j<rows; ++j) {
	// MUST CHANGE THIS!
	delete_matrix(lat_exc_range[i][j].cols,
		      lat_exc_range[i][j].rows,wgt[i][j].lat_exc_w);
	delete_matrix(lat_inh_range[i][j].cols,
		      lat_inh_range[i][j].rows,wgt[i][j].lat_inh_w);
     }
  
  // finally, dealloc the master weight matrix
  delete_matrix(cols,rows,wgt);

  // delete RF range
  for (int i=0; i<max_inp_layers; ++i) delete_matrix(cols,rows,rf_ranges[i]);
  delete [] rf_ranges;

  // dealloc blackboard
  delete net_bb;  
} 

//----------------------------------------------------------------------------
//	alloc_matrix() 
//	  - allocate matrix (or sepcified type, or REAL) and return pointer
//----------------------------------------------------------------------------
REAL **netlayer::alloc_matrix(int cols, int rows) {

  REAL **rp;

  rp = new REAL* [cols];

  for (int i=0; i<cols; ++i) rp[i] = new REAL [rows];

  return rp;

}

//----------------------------------------------------------------------------
// Third argument is just used to specify the return type
NET_RF_Range **netlayer::alloc_matrix(int cols, int rows, 
		NET_RF_Range **rf_range) {

  NET_RF_Range **rfp;

  rfp = new NET_RF_Range* [cols];
  for (int i=0; i<cols; ++i) rfp[i] = new NET_RF_Range[rows];

  return rfp;

}

//----------------------------------------------------------------------------
// Third argument is just used to specify the return type
NET_Weights **netlayer::alloc_matrix(int cols, int rows, NET_Weights **wgts) {

  NET_Weights **wp;

  wp = new NET_Weights* [cols];
  for (int i=0; i<cols; ++i) wp[i] = new NET_Weights [rows];

  return wp;

}

//----------------------------------------------------------------------------
// Third argument is just used to specify the return type
int **netlayer::alloc_matrix(int cols, int rows, int **mat) {

  int **ip;

  ip = new int* [cols];
  for (int i=0; i<cols; ++i) ip[i] = new int [rows];

  return ip;

}

//----------------------------------------------------------------------------
//	delete_matrix() 
//	  - delete matrix (or sepcified type, or REAL) and return pointer
//----------------------------------------------------------------------------
void netlayer::delete_matrix(int cols, int rows, REAL **rp) {

  for (int i=0; i<cols; ++i) delete [] rp[i];
  delete [] rp; 

}

//----------------------------------------------------------------------------
void netlayer::delete_matrix(int cols, int rows, NET_RF_Range **rfp) {

  for (int i=0; i<cols; ++i) delete [] rfp[i];
  delete [] rfp;

}

//----------------------------------------------------------------------------
void netlayer::delete_matrix(int cols, int rows, NET_Weights **wgts) {

  for (int i=0; i<cols; ++i) delete [] wgts[i];
  delete [] wgts;

}

//----------------------------------------------------------------------------
void netlayer::delete_matrix(int cols, int rows, int **ip) {

  for (int i=0; i<cols; ++i) delete [] ip[i];
  delete [] ip;

}

//----------------------------------------------------------------------------
//	alloc_first() 
//	  - allocate as much as possible from info provided by the 
//	    constructor.
//----------------------------------------------------------------------------
void netlayer::alloc_first(int cols, int rows,
                 int max_inp_layers,
                 int exc_rad, int inh_rad) {
 
  // Now, allocate all relevant memory
  // Individual memory allocation within structured components 
  // should also be allocated here.
  //
  //	1. for afferent input parameters, etc.
  //
  netlayer::aff_contris	= new REAL [max_inp_layers];
  netlayer::aff_noises 	= new REAL [max_inp_layers];
  netlayer::alpha_affs	= new REAL [max_inp_layers];
  netlayer::feedings 	= new netlayer* [max_inp_layers];
  netlayer::cogs	= new coord** [max_inp_layers];
  //
  //	2. for the network (matrices)
  //
  netlayer::net_input   = alloc_matrix(cols,rows);
  netlayer::net_aff_input = alloc_matrix(cols,rows);
  netlayer::net_exc_input = alloc_matrix(cols,rows);
  netlayer::net_inh_input = alloc_matrix(cols,rows);
  netlayer::output	= alloc_matrix(cols,rows);
  netlayer::prev_output	= alloc_matrix(cols,rows);
  netlayer::wgt		= alloc_matrix(cols,rows,netlayer::wgt);
  netlayer::wgt_sum 	= alloc_matrix(cols,rows);
  netlayer::or_pref	= alloc_matrix(cols,rows);
  netlayer::or_selectivity = alloc_matrix(cols,rows);

  // 00Nov25 : this needs to be of the size of the retina, not the cortex
  // netlayer::ref_pattern = alloc_matrix(cols,rows);

  //
  // 	3. for the afferent weights (set up a table of poiters first)
  //
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	// don't allocate any memory for afferent inputs yet. Do that in 
	// 	add_inp_layer().
	netlayer::wgt[i][j].aff_ws    = new REAL** [max_inp_layers];
     }
  }

  // add lateral connections : this needs to be done only once.
  add_lat_connections();	

}

//----------------------------------------------------------------------------
//	add_inp_layer() 
//	  - allocate necessary memory and set link to input layer
//----------------------------------------------------------------------------
void netlayer::add_inp_layer(netlayer *feeding, 
	int inp_cols, int inp_rows, int aff_rad) {

  int cur_inp = num_inp_layers;

  // if it is not the first input layer being added, calc_rf_ranges
  // has not been done yet, so do that first.
  if (cur_inp!=0) {

	calc_rf_ranges(cur_inp,cols,rows,inp_cols,inp_rows,aff_rad);
  } 

  if (cur_inp==0) {
  	ref_pattern = alloc_matrix(MAX(inp_cols,cols),MAX(inp_rows,rows));
	cerr << "ref_pattern = " << inp_cols << "x" << inp_rows << endl;
  }

  netlayer::feedings[cur_inp] = feeding;

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// Now allocate memory for the afferent weights
  	wgt[i][j].aff_ws[cur_inp] = alloc_matrix (rf_ranges[cur_inp][i][j].cols,
		rf_ranges[cur_inp][i][j].rows);
     }
  } 

  // allocate center of gravity matrices
  cogs[cur_inp] = new coord* [cols];
  for (int i=0; i<cols; ++i)
        cogs[cur_inp][i] = new coord [rows];

  if (cur_inp!=0) update_blackboard(cur_inp);

  // increment current input count (it becomes 1 after initial add)
  ++num_inp_layers;

}

//----------------------------------------------------------------------------
//	init_blackboard() 
//	  - alloc and fill in the blackboard
//----------------------------------------------------------------------------
void netlayer::init_blackboard(char *name) {

  //	        basic/ lat conn  & aff conn & rf_ranges 
  int max_vars = 100 + cols*rows*(2+2*max_inp_layers);
  cout << "netlayer: max bb vars = " << max_vars << endl;

  // initialize new blackboard
  net_bb = new blackboard (name,max_vars);

  // Add to blackboard : other variables involving iterations will be
  // added in the subroutines called above.
  //
  // - add in the same order as in netlayer.h to enforce consistency
  //   and avoid confusion 
  //
  net_bb->add("cols",&cols);
  net_bb->add("rows",&rows);
  net_bb->add("max_inp_layers",&max_inp_layers);
  net_bb->add("num_inp_layers",&num_inp_layers);

  net_bb->add("aff_rads",aff_rads,max_inp_layers);
  net_bb->add("lat_exc_rad",&lat_exc_rad);
  net_bb->add("lat_inh_rad",&lat_inh_rad);

  net_bb->add("sig_low",&sig_low);
  net_bb->add("sig_high",&sig_high);

  net_bb->add("threshold",&threshold);

  net_bb->add("aff_contris",aff_contris,max_inp_layers);
  net_bb->add("lat_exc_contri",&lat_exc_contri);
  net_bb->add("lat_inh_contri",&lat_inh_contri);

  net_bb->add("aff_noises",aff_noises,max_inp_layers);
  net_bb->add("lat_exc_noise",&lat_exc_noise);
  net_bb->add("lat_inh_noise",&lat_inh_noise);

  net_bb->add("intrinsic_noise",&intrinsic_noise);

  net_bb->add("alpha_affs",alpha_affs,max_inp_layers);
  net_bb->add("alpha_lat_exc",&alpha_lat_exc);
  net_bb->add("alpha_lat_inh",&alpha_lat_inh);

  net_bb->add("net_input",net_input,cols,rows);
  net_bb->add("net_aff_input",net_aff_input,cols,rows);
  net_bb->add("net_exc_input",net_exc_input,cols,rows);
  net_bb->add("net_inh_input",net_inh_input,cols,rows);
  net_bb->add("output",output,cols,rows);
  net_bb->add("wgt_sum",wgt_sum,cols,rows);

  net_bb->add("recursive_lateral_flag",&recursive_lateral_flag);
  net_bb->add("normalization_mode_flag",&normalization_mode_flag);

  // the RF range and lateral ranges are pre-calculated in 
  // init_weights(), so there's no need to save this redundant
  // information.

  // now the interesting part : connection weights.
  char tag[BB_NAME_LEN];
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// 1. the afferent weights
  	for (int k=0; k<num_inp_layers; ++k) {
		sprintf(tag,"wgts[%d][%d].aff_ws[%d]",i,j,k);
		net_bb->add(tag,wgt[i][j].aff_ws[k],
			rf_ranges[k][i][j].cols,rf_ranges[k][i][j].rows);
	}

	// 2. lateral excitatory weights
	sprintf(tag,"wgts[%d][%d].lat_exc_w",i,j);
	net_bb->add(tag,wgt[i][j].lat_exc_w,
			lat_exc_range[i][j].cols,lat_exc_range[i][j].rows);

	// 3. lateral inhibitory weights
	sprintf(tag,"wgts[%d][%d].lat_inh_w",i,j);
	net_bb->add(tag,wgt[i][j].lat_inh_w,
			lat_inh_range[i][j].cols,lat_inh_range[i][j].rows);
     }
  }

  // print warning about link to feedings
  // cout << "netlayer: you have to manually set netlayer::feedings[]\n";
  // cout << "          after load()ing the blackboard image.\n"; 

}


//----------------------------------------------------------------------------
//	add_lat_connections() 
//	  - allocate memory for lateral connections
//----------------------------------------------------------------------------
void netlayer::add_lat_connections() {

  lat_exc_range = alloc_matrix(cols,rows,lat_exc_range);
  lat_inh_range = alloc_matrix(cols,rows,lat_inh_range);

  cerr << "netlayer::add_lat_connections() : adding lat connections - exc_rad="
	<< lat_exc_rad << ", inh_rad=" << lat_inh_rad << endl;

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// 1.1 calculate lateral excitatory ranges
 	lat_exc_range[i][j].lowx = MAX(0,i-lat_exc_rad);
 	lat_exc_range[i][j].lowy = MAX(0,j-lat_exc_rad);
 	lat_exc_range[i][j].highx= MIN(cols-1,i+lat_exc_rad);
 	lat_exc_range[i][j].highy= MIN(rows-1,j+lat_exc_rad);
	lat_exc_range[i][j].cols = lat_exc_range[i][j].highx 
		- lat_exc_range[i][j].lowx + 1;
	lat_exc_range[i][j].rows = lat_exc_range[i][j].highy
	      	- lat_exc_range[i][j].lowy + 1;
 
	// 1.2 Now allocate memory for the lat_exc weights
  	wgt[i][j].lat_exc_w = alloc_matrix(lat_exc_range[i][j].cols,
		lat_exc_range[i][j].rows);

	// 2.1 calculate lateral inhibitory ranges
 	lat_inh_range[i][j].lowx = MAX(0,i-lat_inh_rad);
 	lat_inh_range[i][j].lowy = MAX(0,j-lat_inh_rad);
 	lat_inh_range[i][j].highx= MIN(cols-1,i+lat_inh_rad);
 	lat_inh_range[i][j].highy= MIN(rows-1,j+lat_inh_rad);
	lat_inh_range[i][j].cols = lat_inh_range[i][j].highx 
		- lat_inh_range[i][j].lowx + 1;
	lat_inh_range[i][j].rows = lat_inh_range[i][j].highy
	      	- lat_inh_range[i][j].lowy + 1;
 
	// 2.2 Now allocate memory for the lat_inh weights
  	wgt[i][j].lat_inh_w = alloc_matrix(lat_inh_range[i][j].cols,
		lat_inh_range[i][j].rows);
     }
  } 

}


//----------------------------------------------------------------------------
//	save() 
//	  - save blackboards, using "prefix" 
//----------------------------------------------------------------------------
void netlayer::save(char *prefix) {

  char bbname[BB_NAME_LEN];

  sprintf(bbname,"%s-%s.bbimg",prefix,name);
  net_bb->save(bbname);

}

//----------------------------------------------------------------------------
//	load() 
//	  - load blackboards, using "filename"
//----------------------------------------------------------------------------
void netlayer::load(char *prefix) {

  char bbname[BB_NAME_LEN];

  sprintf(bbname,"%s-%s.bbimg",prefix,name);
  net_bb->load(bbname);

}


//----------------------------------------------------------------------------
//	list() 
//	  - list blackboard
//----------------------------------------------------------------------------
void netlayer::list() {

  net_bb->list();

}


//----------------------------------------------------------------------------
//	act_func() 		    /~~~
//	  - activation function :__/ 
//				   | | 
//			    sig_low  sig_high
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
//	init_weights() 
//	  - initialize lateral and afferent weights
//	  - sets initial weight values to [low,high) 
//----------------------------------------------------------------------------
void netlayer::init_weights(REAL low, REAL high) {

  int lowx, lowy, lowi, lowj;

  cout << "*** Recursive Lateral Flag = " << recursive_lateral_flag << endl;

  // 1. first, the afferent weights 
  srand48(914895);
  for (int i=0; i<cols; ++i) 
     for (int j=0; j<rows; ++j) 

  	for (int k=0; k<num_inp_layers; ++k) { // for all input layers
	
	   for(int x=0; x<rf_ranges[k][i][j].cols; ++x) 
	      for(int y=0; y<rf_ranges[k][i][j].rows; ++y) {
		 /* debug cout << "*** init aff (" << x << "," << y << ") "
			<< rf_ranges[k][i][j].cols << " " 
			<< rf_ranges[k][i][j].rows << endl;
		 */
	         wgt[i][j].aff_ws[k][x][y] = (REAL) drand48()*(high-low)+low;
	      }

	}

  // 2. then, the lateral weights.
  for (int i=0; i<cols; ++i) 
     for (int j=0; j<rows; ++j)  {

	// 2.1 excitatory
	lowx = lat_exc_range[i][j].lowx;
	lowy = lat_exc_range[i][j].lowy;
	for(int x=0; x<lat_exc_range[i][j].cols; ++x) {
	   for(int y=0; y<lat_exc_range[i][j].rows; ++y) {
	      if (recursive_lateral_flag && i==x+lowx && j==y+lowy) {
		wgt[i][j].lat_exc_w[x][y] = -DEATH_FLAG;
	      } else {
	      	wgt[i][j].lat_exc_w[x][y] = (REAL) drand48()*(high-low)+low;
 	      }
	      // make weights symmetric
	      lowi = lat_exc_range[x+lowx][y+lowy].lowx;
	      lowj = lat_exc_range[x+lowx][y+lowy].lowy;
	      wgt[x+lowx][y+lowy].lat_exc_w[i-lowi][j-lowj] 
	  		= wgt[i][j].lat_exc_w[x][y];
	   }
	}

	// 2.2 inhibitory
	lowx = lat_inh_range[i][j].lowx;
	lowy = lat_inh_range[i][j].lowy;
	for(int x=0; x<lat_inh_range[i][j].cols; ++x) {
	   for(int y=0; y<lat_inh_range[i][j].rows; ++y) {
	      if (recursive_lateral_flag && i==x+lowx && j==y+lowy) {
	        wgt[i][j].lat_inh_w[x][y]= -DEATH_FLAG;
	      } else {
	        wgt[i][j].lat_inh_w[x][y]= (REAL) drand48()*(high-low)+low;
	      }
	
	      // make weights symmetric
	      lowi = lat_inh_range[x+lowx][y+lowy].lowx;
	      lowj = lat_inh_range[x+lowx][y+lowy].lowy;
	      wgt[x+lowx][y+lowy].lat_inh_w[i-lowi][j-lowj] 
			= wgt[i][j].lat_inh_w[x][y];
	   }
 	}

	// reset output 
	output[i][j] = 0.0;
	prev_output[i][j] = 0.0;
     }

}

//----------------------------------------------------------------------------
//	normalize_weights() 
//	  - normalize aff and lat weights with sum of weight : outstar
//----------------------------------------------------------------------------
void netlayer::normalize_weights() {

  int topx, topy;
  int lowx, highx, lowy, highy;
  register REAL tmp_wgt;
  register REAL tmp_sum;

  // 1. first, the afferent weights 
  for (int k=0; k<num_inp_layers; ++k) { // for all input layers
     for (int i=0; i<cols; ++i) {
        for (int j=0; j<rows; ++j) {
	
	   REAL sum = 0.0;
	   topx = rf_ranges[k][i][j].cols;
	   topy = rf_ranges[k][i][j].rows;
	   for(int x=0; x<topx; ++x)
	      for(int y=0; y<topy; ++y) {
		 tmp_wgt = wgt[i][j].aff_ws[k][x][y];
	         if (tmp_wgt>0.0) 
		     sum += tmp_wgt;
	      }

	   if (fabs(sum)<=MIN_REAL) sum = 1.0;

	   REAL norm = 1.0 / (sum);
	   for(int x=0; x<topx; ++x)
	      for(int y=0; y<topy; ++y) {
	         if (wgt[i][j].aff_ws[k][x][y]>0.0)
	             wgt[i][j].aff_ws[k][x][y] *= norm;
	      }

	}
     }
  }

  // 2. then, the lateral weights.

  // 2.1 excitatory weights
  // 2.1.1 clear out weight sum
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	wgt_sum[i][j] = 0.0;
     }
  }
  // 2.1.2 : first, sum the weights
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// outstar style normalization : see comment below in the
	// inhibitory weight normalization.
	topx = lat_exc_range[i][j].cols;
	topy = lat_exc_range[i][j].rows;
	for (int x=0; x<topx; ++x) 
	   for(int y=0; y<topy; ++y){
	      tmp_wgt = wgt[i][j].lat_exc_w[x][y];
	      if (tmp_wgt>0.0)
	      	wgt_sum[i][j] += tmp_wgt;
	   }

	if (fabs(wgt_sum[i][j])<=0.0) wgt_sum[i][j] = 1.0;
	else wgt_sum[i][j] = 1.0 / wgt_sum[i][j];
     }
  }
  // 2.1.3 normalize
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// outstar style
	topx = lat_exc_range[i][j].cols;
	topy = lat_exc_range[i][j].rows;
	tmp_sum = wgt_sum[i][j];
	for(int x=0; x<topx; ++x)
	   for(int y=0; y<topy; ++y)
	      if (wgt[i][j].lat_exc_w[x][y]>0.0)
	      	wgt[i][j].lat_exc_w[x][y] *= tmp_sum;

     }
  }

  // 2.2 inhibitory weights
  // 2.2.1 clear out weight sum
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	wgt_sum[i][j] = 0.0;
     }
  }
  // 2.2.2 : first, sum the weights
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// normalize outstar style : this is how it was done in gaus-slissom
	// and it only works this way!
	topx = lat_inh_range[i][j].cols;
	topy = lat_inh_range[i][j].rows;
	for (int x=0; x<topx; ++x) 
	   for(int y=0; y<topy; ++y) {
	      tmp_wgt = wgt[i][j].lat_inh_w[x][y];
	      if (tmp_wgt>0.0)
	      	wgt_sum[i][j] += tmp_wgt;
	   }

	if (fabs(wgt_sum[i][j])<=0.0) wgt_sum[i][j] = 1.0;
	else wgt_sum[i][j] = 1.0 / wgt_sum[i][j];
     }
  }
  // 2.2.3 : then actualy normalize
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	topx = lat_inh_range[i][j].cols;
	topy = lat_inh_range[i][j].rows;
	tmp_sum = wgt_sum[i][j];
	for(int x=0; x<topx; ++x)
	   for(int y=0; y<topy; ++y)
		// normalize outstar style
	      if (wgt[i][j].lat_inh_w[x][y]>0.0)
		wgt[i][j].lat_inh_w[x][y] *= tmp_sum;

     }
  }

}

//----------------------------------------------------------------------------
//	normalize_instar_weights() 
//	  - normalize aff and lat weights with sum of weight : instar style
//	    i.e., normalize incoming weights with the sum of the 
//	    receiving neuron's incoming weights.
//----------------------------------------------------------------------------
void netlayer::normalize_instar_weights() {

  int topx, topy;

  // 1. first, the afferent weights 
  for (int k=0; k<num_inp_layers; ++k) { // for all input layers
     for (int i=0; i<cols; ++i) {
        for (int j=0; j<rows; ++j) {
	
	   REAL sum = 0.0;
	   for(int x=0; x<rf_ranges[k][i][j].cols; ++x)
	      for(int y=0; y<rf_ranges[k][i][j].rows; ++y)
		 sum += (wgt[i][j].aff_ws[k][x][y]);

	   if (fabs(sum)<=MIN_REAL) sum = 1.0;

	   REAL norm = 1.0 / (sum);

	   for(int x=0; x<rf_ranges[k][i][j].cols; ++x)
	      for(int y=0; y<rf_ranges[k][i][j].rows; ++y)
	         wgt[i][j].aff_ws[k][x][y] *= norm;

	}
     }
  }

  // 2. then, the lateral weights.

  // 2.1 excitatory weights
  // 2.1.1 clear out weight sum
  for (int i=0; i<cols; ++i) 
     for (int j=0; j<rows; ++j) 
	wgt_sum[i][j] = 0.0;

  // 2.1.2 : first, sum the weights
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// instar style normalization
	for(int x=lat_exc_range[i][j].lowx;x<=lat_exc_range[i][j].highx; ++x) 
	  for(int y=lat_exc_range[i][j].lowy;y<=lat_exc_range[i][j].highy;++y) 
	     wgt_sum[i][j] += wgt[x][y].lat_exc_w
					[i-lat_exc_range[x][y].lowx]
	 				[j-lat_exc_range[x][y].lowy];

	if (fabs(wgt_sum[i][j])<=0.0) wgt_sum[i][j] = 1.0; 
	else wgt_sum[i][j] = 1.0 / wgt_sum[i][j];
	
     }
  }
  // 2.1.3 normalize
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	int basex = lat_exc_range[i][j].lowx;
	int basey = lat_exc_range[i][j].lowy;
	for(int x=0; x<lat_exc_range[i][j].cols; ++x)
	   for(int y=0; y<lat_exc_range[i][j].rows; ++y)
	      // instar style
	      wgt[i][j].lat_exc_w[x][y] *= wgt_sum[x+basex][y+basey];

	/*
	if (output[i][j]>0.0) {
	for(int x=lat_exc_range[i][j].lowx; x<=lat_exc_range[i][j].highx; ++x) 
	  for(int y=lat_exc_range[i][j].lowy;y<=lat_exc_range[i][j].highy;++y)
	     wgt[x][y].lat_exc_w
		[i-lat_exc_range[x][y].lowx]
		[j-lat_exc_range[x][y].lowy]
	     *= wgt_sum[i][j];
	}
	*/

     }
  }

  // 2.2 inhibitory weights
  // 2.2.1 clear out weight sum
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	wgt_sum[i][j] = 0.0;
     }
  }
  // 2.2.2 : first, sum the weights
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// normalize instar style
	for(int x=lat_inh_range[i][j].lowx;x<=lat_inh_range[i][j].highx; ++x) 
	  for(int y=lat_inh_range[i][j].lowy;y<=lat_inh_range[i][j].highy;++y) 
	     wgt_sum[i][j] += wgt[x][y].lat_inh_w
					[i-lat_inh_range[x][y].lowx]
					[j-lat_inh_range[x][y].lowy];

	if (fabs(wgt_sum[i][j])<=0.0) wgt_sum[i][j] = 1.0;
	else wgt_sum[i][j] = 1.0 / wgt_sum[i][j];

     }
  }
  // 2.2.3 : then actualy normalize
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	int basex = lat_inh_range[i][j].lowx;
	int basey = lat_inh_range[i][j].lowy;
	for(int x=0; x<lat_inh_range[i][j].cols; ++x)
	   for(int y=0; y<lat_inh_range[i][j].rows; ++y)
		// normalize instar style
	       	wgt[i][j].lat_inh_w[x][y] *= wgt_sum[x+basex][y+basey];
	/* it can be done this way as well, but there's no 
	   immediate speedup
	if (output[i][j]>0.0) {
	for(int x=lat_inh_range[i][j].lowx; x<=lat_inh_range[i][j].highx; ++x) 
	  for(int y=lat_inh_range[i][j].lowy;y<=lat_inh_range[i][j].highy;++y)
	     wgt[x][y].lat_inh_w
		[i-lat_inh_range[x][y].lowx]
		[j-lat_inh_range[x][y].lowy]
	     *= wgt_sum[i][j];
	}
	*/
     }
  }

}

//----------------------------------------------------------------------------
//	calc_rf_ranges() 
//	  - calculate afferent RF ranges and lateral ranges
//----------------------------------------------------------------------------
void netlayer::calc_rf_ranges(int cur_inp, 
		int cols, int rows, 
		int inp_cols, int inp_rows,
 		int aff_rad){

  int unpadded_cols, unpadded_rows; 
  int col_offset, row_offset;

  if ((padded_flag==1 && cur_inp == 0) || (padded_flag>=2)) {
  //   ------------------------------      -------------- 
  //	01Jul06
  //     default for 01jun16 -- sims	   00nov default      
  //	 * pad the RETINA only.		   * pad both retina and
  //     * do not pad intracolumnar 	   * cortex
  //					   * old sims need to be updated
  //					     where the value was previously 1
	cerr << "================= padded_flag = " << padded_flag << endl;
  	unpadded_cols = inp_cols - 2 * aff_rad;
  	unpadded_rows = inp_rows - 2 * aff_rad;
  	col_offset = aff_rad;
  	row_offset = aff_rad;
  } else {
	cerr << "================= padded_flag = 0  " << endl;
	unpadded_cols = inp_cols;
	unpadded_rows = inp_rows;
  	col_offset = 0;
  	row_offset = 0;
  }


  // cout << "calc_rf_ranges: " << cur_inp << endl;
  // Before going to the next step, alloc and calculate RF ranges
  //
  // 	1. allocate memory need RF ranges for [cols][rows]
  //
  if (inp_cols==0 || inp_rows==0) { // in case there's no input layer
	return;
  } else {
  	rf_ranges[cur_inp] = alloc_matrix(cols,rows,rf_ranges[cur_inp]);
  }
  //
  //	2. calculate the RF ranges for the respective [cols][rows]
  //		- must deal with different sized input vs. current layer
  //
  REAL mag_x, mag_y;

  mag_x = (REAL)cols/(REAL)unpadded_cols; 
  mag_y = (REAL)rows/(REAL)unpadded_rows;

  aff_rads[cur_inp] = aff_rad;

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

  	int cent_x = (int)(i/mag_x)+col_offset;
	int cent_y = (int)(j/mag_y)+row_offset;
   	// printf("cent(%d,%d) : mag(%f,%f) :",cent_x,cent_y,mag_x,mag_y);
 	rf_ranges[cur_inp][i][j].lowx = MAX(0,cent_x-aff_rad);
 	rf_ranges[cur_inp][i][j].lowy = MAX(0,cent_y-aff_rad);
 	rf_ranges[cur_inp][i][j].highx= MIN(inp_cols-1,cent_x+aff_rad);
 	rf_ranges[cur_inp][i][j].highy= MIN(inp_rows-1,cent_y+aff_rad);
	rf_ranges[cur_inp][i][j].cols = rf_ranges[cur_inp][i][j].highx 
		- rf_ranges[cur_inp][i][j].lowx + 1;
	rf_ranges[cur_inp][i][j].rows = rf_ranges[cur_inp][i][j].highy
	      - rf_ranges[cur_inp][i][j].lowy + 1;

	/* debug
	cout << "*** ranges : " << i << " " << j 
		<< " " << rf_ranges[cur_inp][i][j].lowx
		<< " " << rf_ranges[cur_inp][i][j].lowy
		<< " " << rf_ranges[cur_inp][i][j].highx
		<< " " << rf_ranges[cur_inp][i][j].highy << endl;
	*/

     }
  }

}

//----------------------------------------------------------------------------
//	update_blackboard() 
//	  - update blackboard after additional input layer was added
//	    with add_inp_layer()
//----------------------------------------------------------------------------
void netlayer::update_blackboard(int cur_inp) {

  char tag[BB_NAME_LEN];

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

        // 1. the afferent weights
        sprintf(tag,"wgts[%d][%d].aff_ws[%d]",i,j,cur_inp);
        net_bb->add(tag,wgt[i][j].aff_ws[cur_inp],
               rf_ranges[cur_inp][i][j].cols,rf_ranges[cur_inp][i][j].rows);

     }
  }

}


//----------------------------------------------------------------------------
//	activate() 
//	  - activate map : output[][] is the final activity
//----------------------------------------------------------------------------
void netlayer::activate(int calc_aff) {

  //
  // 1. Calculate netinput
  //
  // 	calls calc_affresp() or calc_latresp()
  //	(calc_netinput() is used here only for compatibility with
  //	 previously written test programs in ./test)
  //
  calc_netinput(calc_aff);

  //
  // 2. Calculate output (by applying the activation function)
  //

  apply_actfunc();

}


//----------------------------------------------------------------------------
//	calc_netinput() 
//	  - calculate netinput
//	  - todo : separate into different functions
//----------------------------------------------------------------------------
void netlayer::calc_netinput(int calc_aff) {

  // Call appropriate subroutine 
  if (calc_aff) {
	calc_affresp();
	calc_contrib_sum(calc_aff);
  } else {
	//reset_matrix(net_input,cols,rows);
	// leave net_aff_input intact!
	calc_latresp();
	calc_contrib_sum(calc_aff);
  }
}
  
//----------------------------------------------------------------------------
//	calc_affresp() 
//	  - calculate afferent contribution only
//----------------------------------------------------------------------------
void netlayer::calc_affresp() {

  REAL	aff_netinp;
  int	basex, basey;
  REAL	**tmp_wgt_pp;
  REAL	**tmp_output_pp;
  REAL  tmp_wgt;
  
  int	highx,highy;

  cerr << "calcaffresp" << endl;
  //
  // 1. For all afferent weights of unit [i][j]
  //
  reset_matrix(net_aff_input,cols,rows);
  for (int k=0; k<num_inp_layers; ++k) {
     tmp_output_pp = feedings[k]->output;
     for (int i=0; i<cols; ++i) {
        for (int j=0; j<rows; ++j) {
           aff_netinp 	= 0.0; 
  	   basex 	= rf_ranges[k][i][j].lowx;
	   basey 	= rf_ranges[k][i][j].lowy;
	   highx	= rf_ranges[k][i][j].highx;
	   highy	= rf_ranges[k][i][j].highy;
     	   tmp_wgt_pp 	= wgt[i][j].aff_ws[k];
	   for (int x=basex; x<=highx; ++x)	
	      for (int y=basey; y<=highy; ++y)	{
			tmp_wgt = tmp_wgt_pp[x-basex][y-basey];		
		        if (tmp_wgt>0.0) {
	   		   aff_netinp += tmp_output_pp[x][y] * tmp_wgt;
			}
	      }

	   net_aff_input[i][j] += aff_contris[k] * aff_netinp;
	 }
     }
     // todo: must consider what to do when there are more than
     //  	one input layer. 9/26/1998
  } 

  /* 2000jan27
  // scale it
  REAL divisor = 1.0/(REAL)num_inp_layers;
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	net_aff_input[i][j] *= divisor;
     }
  }
  */

	
  // 1. done

}

//----------------------------------------------------------------------------
//	calc_affresp_padded() 
//	  - calculate afferent contribution only
//	  - assume retina is padded, so don't need to reference the
//	    ranges.
//----------------------------------------------------------------------------
void netlayer::calc_affresp_padded() {

  REAL	aff_netinp;
  int	basex, basey;
  int	rfcols, rfrows;
  REAL	**tmp_wgt_pp;
  REAL	**tmp_output_pp;
  REAL  tmp_wgt;
  int	xoffset,yoffset;
  int	xloc;

  cerr << "calc_affresp_padded()" << endl;
  //
  // 1. For all afferent weights of unit [i][j]
  //
  reset_matrix(net_aff_input,cols,rows);
  for (int k=0; k<num_inp_layers; ++k) {
     rfcols = rfrows = 2*aff_rads[k]+1;
     tmp_output_pp = feedings[k]->output;
     // REAL **feed_mat = feedings[k]->output;
     for (int i=0; i<cols; ++i) {
        for (int j=0; j<rows; ++j) {
           aff_netinp = 0.0; 
	   // REAL **wgt_mat = wgt[i][j].aff_ws[k];
	   xoffset = rf_ranges[k][i][j].lowx;
	   yoffset = rf_ranges[k][i][j].lowy;
     	   tmp_wgt_pp 	= wgt[i][j].aff_ws[k];
	   for (int x=0; x<rfcols; ++x) {
	      xloc = x+xoffset;
	      for (int y=0; y<rfrows; ++y) {
			tmp_wgt = tmp_wgt_pp[x][y];		
		        if (tmp_wgt>0.0) {
	   		   aff_netinp += tmp_output_pp[xloc][y+yoffset]*tmp_wgt;
			}
	      }
	      
	   }

	   net_aff_input[i][j] += aff_contris[k] * aff_netinp;
	 }
     }
     // todo: must consider what to do when there are more than
     //  	one input layer. 9/26/1998
  } 

  /* 2000jan27
  // scale it
  REAL divisor = 1.0/(REAL)num_inp_layers;
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	net_aff_input[i][j] *= divisor;
     }
  }
  */
	
  // 1. done

}


//----------------------------------------------------------------------------
//	calc_latresp_optimized() 
//	  - calculate lateral response only : optimized
//----------------------------------------------------------------------------
void netlayer::calc_latresp_optimized() {

  REAL	exc_netinp, inh_netinp;
  int	basex, basey;

  reset_matrix(net_exc_input,cols,rows);
  reset_matrix(net_inh_input,cols,rows);

  //
  // 1. Calculate lateral contribution
  //
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
   
	// Condition 1. For active units only
        if (output[i][j]>0.0) {
   
           // 1.1 For all excitatory lateral connections
           // 	later than: v1.12 use outstar calculation here.
   
  	   basex = lat_exc_range[i][j].lowx;
	   basey = lat_exc_range[i][j].lowy;
	   exc_netinp = 0.0;
	   REAL **wgt_mat = wgt[i][j].lat_exc_w;
	   int sx = lat_exc_range[i][j].lowx;
	   int ex = lat_exc_range[i][j].highx;
           int sy = lat_exc_range[i][j].lowy;
	   int ey = lat_exc_range[i][j].highy;
           for (int x=sx; x<=ex; ++x) {
	      int xloc = x-basex;
              for (int y=sy; y<=ey; ++y) {
		   // outstar
		   if (wgt_mat[xloc][y-basey]>0.0) {
		      net_exc_input[x][y] += output[i][j] 
			   * wgt_mat[xloc][y-basey];
		   }
	      }
	   } 
	   // 1.1 done
   
     	   // 1.2 For all inhibitory lateral connections
     	   // 	later than: v1.12 use outstar calculation here.
  	   basex = lat_inh_range[i][j].lowx;
	   basey = lat_inh_range[i][j].lowy;
	   inh_netinp = 0.0;
	   wgt_mat = wgt[i][j].lat_inh_w;
	   sx = lat_inh_range[i][j].lowx;
	   ex = lat_inh_range[i][j].highx;
           sy = lat_inh_range[i][j].lowy;
	   ey = lat_inh_range[i][j].highy;
           for (int x=sx; x<=ex; ++x) {
	      int xloc = x-basex;
              for (int y=sy; y<=ey; ++y) {
		   // outstar
		   if (wgt_mat[xloc][y-basey]>0.0) {
		      net_inh_input[x][y] += output[i][j] 
			   * wgt_mat[xloc][y-basey];
		   }
	      }
 	   }
	   // 1.2 done
	} // Condition 1. end

     } // end of for j
  } // end of for i
  // 1. done 
}
   
//----------------------------------------------------------------------------
//	calc_latresp() 
//	  - calculate lateral response only
//----------------------------------------------------------------------------
void netlayer::calc_latresp() {

	calc_latresp_exc(output);
	calc_latresp_inh(output);

}

//----------------------------------------------------------------------------
void netlayer::calc_latresp_exc(REAL **output) {

  REAL	exc_netinp;
  int	basex, basey;
  int	lowx,highx,lowy,highy;
  REAL	tmp_wgt;
  REAL	tmp_out;
  REAL  **tmp_wgt_pp;

  reset_matrix(net_exc_input,cols,rows);

  //
  // 1. Calculate lateral contribution
  //
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
   
	// Condition 1. For active units only
        if (output[i][j]>0.0) {
   
           // 1.1 For all excitatory lateral connections
           // 	later than: v1.12 use outstar calculation here.
   
	   exc_netinp 	= 0.0;
	   lowx 	= lat_exc_range[i][j].lowx;
	   highx 	= lat_exc_range[i][j].highx;
	   lowy 	= lat_exc_range[i][j].lowy;
	   highy 	= lat_exc_range[i][j].highy;

	   tmp_out = output[i][j];
	   tmp_wgt_pp = wgt[i][j].lat_exc_w;
           for (int x=lowx; x<=highx; ++x) {
              for (int y=lowy; y<=highy; ++y) {
		   // outstar
		   tmp_wgt = tmp_wgt_pp[x-lowx][y-lowy];	
		   if (tmp_wgt>0.0) 
		      net_exc_input[x][y] += tmp_out * tmp_wgt;
	      }
	   } 
	   // 1.1 done
   
	} // Condition 1. end

     } // end of for j
  } // end of for i
  // 1. done 

}
//----------------------------------------------------------------------------
void netlayer::calc_latresp_inh(REAL **output) {

  REAL	inh_netinp;
  int	basex, basey;
  int	lowx,highx,lowy,highy;
  register REAL	tmp_wgt;
  register REAL	tmp_out;
  REAL  **tmp_wgt_pp;

  reset_matrix(net_inh_input,cols,rows);

  //
  // 1. Calculate lateral contribution
  //
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
   
	// Condition 1. For active units only
        if (output[i][j]>0.0) {
   
     	   // 1.1 For all inhibitory lateral connections
     	   // 	later than: v1.12 use outstar calculation here.
	   inh_netinp = 0.0;
	   lowx 	= lat_inh_range[i][j].lowx;
	   highx 	= lat_inh_range[i][j].highx;
	   lowy 	= lat_inh_range[i][j].lowy;
	   highy 	= lat_inh_range[i][j].highy;
	   tmp_out	= output[i][j];
	   tmp_wgt_pp	= wgt[i][j].lat_inh_w;
           for (int x=lowx; x<=highx; ++x) {
              for (int y=lowy; y<=highy; ++y) {
		   // outstar
		   tmp_wgt = tmp_wgt_pp[x-lowx][y-lowy];
		   if (tmp_wgt>0.0) 
			net_inh_input[x][y] += tmp_out * tmp_wgt;
	      }
 	   }
	   // 1.2 done
	} // Condition 1. end

     } // end of for j
  } // end of for i
  // 1. done 

}
//----------------------------------------------------------------------------
//	calc_contrib_sum() 
//	  - calculate net_input from aff and lateral contribs
//----------------------------------------------------------------------------
void netlayer::calc_contrib_sum(int calc_aff){
  //
  // Calculate net_input from calculated aff(with calc_affresp()) and 
  // lat responses (calculated right above).
  //

  reset_matrix(net_input,cols,rows);

  if (calc_aff) {

     	for (int i=0; i<cols; ++i) {
     	   for (int j=0; j<rows; ++j) {
		net_input[i][j] = net_aff_input[i][j];
     	   }
  	}

  } else {

     	for (int i=0; i<cols; ++i) {
     	   for (int j=0; j<rows; ++j) {
		// net_aff_input[i][j] has already been multiplied
		// by aff_contris[]
		net_input[i][j] = net_aff_input[i][j] 
		   + lat_exc_contri * net_exc_input[i][j] 
		   - lat_inh_contri * net_inh_input[i][j];
     	   }
  	}
  }
}

//----------------------------------------------------------------------------
//	set_parameter() 
//	  - set parameter through a command language-like text interface
//	  - this is temporary
//	  - command syntax :  set <param_name>{[<col>]{[<row>]}} = <value>
//		'{ }' : optional stuff
//----------------------------------------------------------------------------
void netlayer::set_parameter(char *cmd) {
   
   net_bb->set_parameter(cmd);

   bookkeep_parameters();
}
//----------------------------------------------------------------------------
// when the parser was passed
void netlayer::set_parameter(parser *sim_parser) {
   
   net_bb->set_parameter(sim_parser);

   bookkeep_parameters();
}
//----------------------------------------------------------------------------
// just take the param name and value
void netlayer::set_parameter(char *param_name, REAL value) {

  parser *cmdp = new parser(param_name);

  cmdp->match_param(param_name,&cols,&rows);

  int param_idx = net_bb->force_find(param_name);

  if (param_idx < 0) {
	cout << "netlayer::set_parameter() : no such parameter " << param_name
	     << endl;
 	return;
  }

  switch (net_bb->table[param_idx].type) {
      case BB_INT: *(net_bb->table[param_idx].data.i) = (int)value; break;
      case BB_IP: net_bb->table[param_idx].data.ip[cols] = (int)value; break;
      case BB_IPP: net_bb->table[param_idx].data.ipp[cols][rows] = (int)value; 
		break;
      case BB_REAL: *(net_bb->table[param_idx].data.r) = value; break;
      case BB_RP: net_bb->table[param_idx].data.rp[cols] = value; break;
      case BB_RPP: net_bb->table[param_idx].data.rpp[cols][rows] = value; break;
  }

  delete cmdp;

  bookkeep_parameters();
}
//----------------------------------------------------------------------------
// take base param name and column as an argument 
void netlayer::set_parameter(char *param_name, int col, REAL val) {
	
  char cmd_str[BB_NAME_LEN];

  sprintf(cmd_str,"set %s[%d] = %f",param_name,col,val);
  
  set_parameter(cmd_str);

  bookkeep_parameters();

}
  
//----------------------------------------------------------------------------
// take base param name and column/row as an argument 
void netlayer::set_parameter(char *param_name, int col, int row, REAL val) {
	
  char cmd_str[BB_NAME_LEN];

  sprintf(cmd_str,"set %s[%d][%d] = %f",param_name,col,row,val);
  
  set_parameter(cmd_str);

}
  

//----------------------------------------------------------------------------
//	learn() 
//	  - using delta(), modify weights.
//----------------------------------------------------------------------------
void netlayer::learn() {

  int basex, basey, basei, basej;

  register REAL tmp_wgt;
  register REAL **tmp_wgt_pp;
  register int  tmp_cols, tmp_rows;

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

  // 1.1 For all afferent weights of unit [i][j]

        for (int k=0; k<num_inp_layers; ++k) {
	   tmp_wgt_pp 	= wgt[i][j].aff_ws[k];
	   basex 	= rf_ranges[k][i][j].lowx;
	   basey 	= rf_ranges[k][i][j].lowy;
	   tmp_cols 	=  rf_ranges[k][i][j].cols;
	   tmp_rows	=  rf_ranges[k][i][j].rows;
	   for (int x=0; x<tmp_cols; ++x)	
	      for (int y=0; y<tmp_rows; ++y)	
		 if (tmp_wgt_pp[x][y]>0.0) 
	             tmp_wgt_pp[x][y] += delta_aff(i,j,k,x+basex,y+basey);
	}
	
	// 1.1 done
	

        if (output[i][j]>0.0) {

  	        // 1.2 For all excitatory lateral connections
		tmp_wgt_pp 	= wgt[i][j].lat_exc_w;
		basex = lat_exc_range[i][j].lowx;
		basey = lat_exc_range[i][j].lowy;
		tmp_cols	= lat_exc_range[i][j].cols;
		tmp_rows	= lat_exc_range[i][j].rows;	
		for (int x=0; x<tmp_cols; ++x) 
		   for (int y=0; y<tmp_rows; ++y)
		      if (tmp_wgt_pp[x][y]>0.0) 
			 tmp_wgt_pp[x][y] += delta_lat(i,j,x+basex,y+basey,
							 alpha_lat_exc);
		// 1.2 done

	  	// 1.3 For all inhibitory lateral connections

	       if (alpha_lat_inh>0.000001) {
		tmp_wgt_pp 	= wgt[i][j].lat_inh_w;
		basex = lat_inh_range[i][j].lowx;
		basey = lat_inh_range[i][j].lowy;
		tmp_cols	= lat_inh_range[i][j].cols;
		tmp_rows	= lat_inh_range[i][j].rows;	
		for (int x=0; x<tmp_cols; ++x) 
		   for (int y=0; y<tmp_rows; ++y)     
		      if (tmp_wgt_pp[x][y]>0.0) 
			 tmp_wgt_pp[x][y] += delta_lat(i,j,x+basex,y+basey,
							alpha_lat_inh);
		}
		// 1.3 done
        }
     }

  } // 1. done : all weights were modified -- it has yet to be normalized.

}

void netlayer::learn_v() {

  netlayer::learn();

}


//----------------------------------------------------------------------------
//	delta() 
//	  - calculate delta, on which weight adaptation is based on
//----------------------------------------------------------------------------
inline REAL netlayer::delta_aff(int col, int row, int cur, int x, int y) {

// for afferent weights : the current input layer is specified by 'cur_inp'

  return (alpha_affs[cur] * (output[col][row] * feedings[cur]->output[x][y]));

}
//----------------------------------------------------------------------------
inline REAL netlayer::delta_lat(int col, int row, int x, int y, REAL alpha) {

// for lateral weights

  return (alpha * (output[col][row]*output[x][y]));

}

//----------------------------------------------------------------------------
//	get_parameter() 
//	  - get parameter value : always return in REAL. Only supports
//	    scalar variables. Returns 0.0 in all other cases.
//----------------------------------------------------------------------------
REAL netlayer::get_parameter(char *param) {

  char param_name[BB_NAME_LEN]; 
  int  cols, rows;
  int param_idx;

  parser *cmdp = new parser(param);

  cmdp->match_param(param_name,&cols,&rows);

  param_idx = net_bb->find(param_name);

  switch (net_bb->table[param_idx].type) {
      case BB_INT: return (REAL)(*(net_bb->table[param_idx].data.i));
      case BB_IP: return (REAL)(net_bb->table[param_idx].data.ip[cols]);
      case BB_IPP: return (REAL)(net_bb->table[param_idx].data.ipp[cols][rows]);
      case BB_REAL: return *(net_bb->table[param_idx].data.r);
      case BB_RP: return net_bb->table[param_idx].data.rp[cols];
      case BB_RPP: return net_bb->table[param_idx].data.rpp[cols][rows];
  }

  delete cmdp;

  return 0.0;

}
//----------------------------------------------------------------------------
REAL netlayer::get_parameter(char *param, int col) {
	
  char param_name[BB_NAME_LEN];

  sprintf(param_name,"%s[%d]",param,col);
  
  return get_parameter(param_name);

}
//----------------------------------------------------------------------------
REAL netlayer::get_parameter(char *param, int col, int row) {
	
  char param_name[BB_NAME_LEN];

  sprintf(param_name,"%s[%d][%d]",param,col,row);
  
  return get_parameter(param_name);

}


//----------------------------------------------------------------------------
//	reset_matrix() 
//	  - reset matrix to all zeros
//----------------------------------------------------------------------------
void netlayer::reset_matrix(int **mat, int cols, int rows) {

  for (int x=0; x<cols; ++x)
    for (int y=0; y<rows; ++y)
	mat[x][y] = 0;
    
}

//----------------------------------------------------------------------------
void netlayer::reset_matrix(REAL **mat, int cols, int rows) {

  for (int x=0; x<cols; ++x)
    for (int y=0; y<rows; ++y)
	mat[x][y] = 0.0;

}


//----------------------------------------------------------------------------
//	reset_output() 
//	  - reset output activity
//----------------------------------------------------------------------------
void netlayer::reset_output() {

   reset_matrix(output,cols,rows);
   // reset_matrix(net_input,cols,rows);

}


//----------------------------------------------------------------------------
//	resize_lat() 
//	  - resize lateral exc range
//----------------------------------------------------------------------------
void netlayer::resize_lat_exc_rad(int new_rad) {

  // 1. calculate range and allocate new, while copying over
  //	old weights

  if (new_rad == lat_exc_rad) return;
  if (new_rad > lat_exc_rad) {
	warning("netlayer::resize_lat_exc_rad",
	  "resizing to larger exc radius not allowed");
	return;
  }

  NET_RF_Range **lat_exc_range = alloc_matrix(cols,rows,lat_exc_range);
  REAL **temp_exc_w;

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// 1.1 calculate lateral excitatory ranges
 	lat_exc_range[i][j].lowx = MAX(0,i-new_rad);
 	lat_exc_range[i][j].lowy = MAX(0,j-new_rad);
 	lat_exc_range[i][j].highx= MIN(cols-1,i+new_rad);
 	lat_exc_range[i][j].highy= MIN(rows-1,j+new_rad);
	lat_exc_range[i][j].cols = lat_exc_range[i][j].highx 
		- lat_exc_range[i][j].lowx + 1;
	lat_exc_range[i][j].rows = lat_exc_range[i][j].highy
	      	- lat_exc_range[i][j].lowy + 1;
 
	// 1.2 Now allocate memory for the lat_exc weights (in a temp
	//	variable temp_exc_w.
	temp_exc_w = alloc_matrix(lat_exc_range[i][j].cols,
                lat_exc_range[i][j].rows);

    	// 1.3 Clear new one's lat_exc weights
	for (int x=0; x<lat_exc_range[i][j].cols; ++x) 
	   for (int y=0; y<lat_exc_range[i][j].rows; ++y)
		temp_exc_w[x][y] = 0.0;
		
 	// 1.4 Now copy over the old weights to the new wgt matrix
 	int new_offset_x=0, new_offset_y=0, old_offset_x=0, old_offset_y=0;
 	if (lat_exc_rad > new_rad) {
	  // old_offset = lat_exc_rad-new_rad; new_offset = 0;
	  old_offset_x = lat_exc_range[i][j].lowx 
		       - netlayer::lat_exc_range[i][j].lowx;
	  old_offset_y = lat_exc_range[i][j].lowy 
		       - netlayer::lat_exc_range[i][j].lowy;
	  new_offset_x = 0;
	  new_offset_y = 0;
	} else { 
	  old_offset_x = old_offset_y = 0; 
	  new_offset_x = new_rad-lat_exc_rad;
	  new_offset_y = new_rad-lat_exc_rad;
	  
	}
	for (int x=0; x<lat_exc_range[i][j].cols; ++x) 
	   for (int y=0; y<lat_exc_range[i][j].rows; ++y) 
	   	temp_exc_w[x+new_offset_x][y+new_offset_y] 
			= wgt[i][j].lat_exc_w[x+old_offset_x][y+old_offset_y];

	// 1.5 free the old weights and link it to the new one
	delete_matrix(netlayer::lat_exc_range[i][j].cols,
		netlayer::lat_exc_range[i][j].rows,wgt[i][j].lat_exc_w);
	wgt[i][j].lat_exc_w = temp_exc_w;

	// 1.7 update the blackboard (no need to free anything here) 
	char tag[BB_NAME_LEN];
	sprintf(tag,"wgts[%d][%d].lat_exc_w",i,j);
	net_bb->update(tag,temp_exc_w,
			lat_exc_range[i][j].cols, lat_exc_range[i][j].rows);
		
     }
  } 

  // 2. Free the old ranges and link it up to the new one
  delete_matrix(cols,rows,netlayer::lat_exc_range);
  netlayer::lat_exc_range = lat_exc_range;

  // 3. update lat_exc_rad
  netlayer::lat_exc_rad = new_rad;

}
//----------------------------------------------------------------------------
// external interface for use with simenv class
int netlayer::call_resize_lat_exc_rad(int argc, char *argv[]) {

  if (argc!=1) return 0;

  net_bb->build_hash();
  resize_lat_exc_rad(atoi(argv[1]));
  net_bb->destroy_hash();

  return 1;

}

//----------------------------------------------------------------------------
//	resize_inh_lat() 
//	  - resize lateral inh range
//----------------------------------------------------------------------------
void netlayer::resize_lat_inh_rad(int new_rad) {

  // 1. calculate range and allocate new, while copying over
  //	old weights

  if (new_rad == lat_inh_rad) return;
  if (new_rad > lat_inh_rad) {
	warning("netlayer::resize_lat_inh_rad",
	  "resizing to larger inh radius not allowed");
	return;
  }

  NET_RF_Range **lat_inh_range = alloc_matrix(cols,rows,lat_inh_range);
  REAL **temp_inh_w;

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// 1.1 calculate lateral inhitatory ranges
 	lat_inh_range[i][j].lowx = MAX(0,i-new_rad);
 	lat_inh_range[i][j].lowy = MAX(0,j-new_rad);
 	lat_inh_range[i][j].highx= MIN(cols-1,i+new_rad);
 	lat_inh_range[i][j].highy= MIN(rows-1,j+new_rad);
	lat_inh_range[i][j].cols = lat_inh_range[i][j].highx 
		- lat_inh_range[i][j].lowx + 1;
	lat_inh_range[i][j].rows = lat_inh_range[i][j].highy
	      	- lat_inh_range[i][j].lowy + 1;
 
	// 1.2 Now allocate memory for the lat_inh weights (in a temp
	//	variable temp_inh_w.
	temp_inh_w = alloc_matrix(lat_inh_range[i][j].cols,
                lat_inh_range[i][j].rows);

    	// 1.3 Clear new one's lat_inh weights
	for (int x=0; x<lat_inh_range[i][j].cols; ++x) 
	   for (int y=0; y<lat_inh_range[i][j].rows; ++y)
		temp_inh_w[x][y] = 0.0;
		
 	// 1.4 Now copy over the old weights to the new wgt matrix
 	int new_offset_x, new_offset_y, old_offset_x, old_offset_y;
 	if (lat_inh_rad > new_rad) {
	  // old_offset = lat_inh_rad-new_rad; new_offset = 0;
	  old_offset_x = lat_inh_range[i][j].lowx 
		       - netlayer::lat_inh_range[i][j].lowx;
	  old_offset_y = lat_inh_range[i][j].lowy 
		       - netlayer::lat_inh_range[i][j].lowy;
	  new_offset_x = 0;
	  new_offset_y = 0;
	} else { 
	  old_offset_x = old_offset_y = 0; 
	  new_offset_x = new_rad-lat_inh_rad;
	  new_offset_y = new_rad-lat_inh_rad;
	  
	}
	for (int x=0; x<lat_inh_range[i][j].cols; ++x) 
	   for (int y=0; y<lat_inh_range[i][j].rows; ++y) 
	   	temp_inh_w[x+new_offset_x][y+new_offset_y] 
			= wgt[i][j].lat_inh_w[x+old_offset_x][y+old_offset_y];

	// 1.5 free the old weights and link it to the new one
	delete_matrix(netlayer::lat_inh_range[i][j].cols,
		netlayer::lat_inh_range[i][j].rows,wgt[i][j].lat_inh_w);
	wgt[i][j].lat_inh_w = temp_inh_w;

	// 1.7 update the blackboard (no need to free anything here) 
	char tag[BB_NAME_LEN];
	sprintf(tag,"wgts[%d][%d].lat_inh_w",i,j);
	net_bb->update(tag,temp_inh_w,
			lat_inh_range[i][j].cols, lat_inh_range[i][j].rows);
		
     }
  } 

  // 2. Free the old ranges and link it up to the new one
  delete_matrix(cols,rows,netlayer::lat_inh_range);
  netlayer::lat_inh_range = lat_inh_range;

  // 3. update lat_inh_rad
  netlayer::lat_inh_rad = new_rad;

}
//----------------------------------------------------------------------------
// external interface for use with simenv class
int netlayer::call_resize_lat_inh_rad(int argc, char *argv[]) {

  if (argc!=1) return 0;

  net_bb->build_hash();
  resize_lat_inh_rad(atoi(argv[1]));
  net_bb->destroy_hash();

  return 1;

}

//----------------------------------------------------------------------------
//	kill_lat_weights() 
//	  - kill weak connections
//----------------------------------------------------------------------------
void  netlayer::kill_lat_weights(REAL threshold) {

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	// 1. for all exc weights 
	for (int x=0; x<lat_exc_range[i][j].cols; ++x) {
	   for (int y=0; y<lat_exc_range[i][j].rows; ++y) {
		if (wgt[i][j].lat_exc_w[x][y] <= threshold) {
		   wgt[i][j].lat_exc_w[x][y] = -DEATH_FLAG;
		}
	   }
	}
	/* 2000 jan 18 : removed 
	// 2. for all inh weights
	for (int x=0; x<lat_inh_range[i][j].cols; ++x) {
	   for (int y=0; y<lat_inh_range[i][j].rows; ++y) {
		if (wgt[i][j].lat_inh_w[x][y] <= threshold) {
		   wgt[i][j].lat_inh_w[x][y] = -DEATH_FLAG;
		}
	   }
	}
 	*/
	// 2. End
     }
  }

}
//----------------------------------------------------------------------------
// external interface for use with simenv class
int netlayer::call_kill_lat_weights(int argc, char *argv[]) {

  if (argc!=1) return 0;
  
  kill_lat_weights(atof(argv[1]));

  return 1;

}


//----------------------------------------------------------------------------
//	update_sig() 
//	  - update sig_high, sig_low, and sig_diff
//----------------------------------------------------------------------------
void netlayer::update_sig(REAL low, REAL high) {

  sig_high = high;
  sig_low = low;

  bookkeep_parameters();

}
//----------------------------------------------------------------------------
// external interface for use with simenv class
int netlayer::call_update_sig(int argc, char *argv[]) {

  if (argc!=2) return 0;

  update_sig(atof(argv[1]),atof(argv[2]));

  return 1;

}
  
//----------------------------------------------------------------------------
//	bookkeep_parameters() 
//	  - keep dependancies up-to-date
//----------------------------------------------------------------------------
void netlayer::bookkeep_parameters() {

  // Put actions to be taken after certain variables have been 
  //	modified.

  // 1. This is the only one for now.
  sig_diff = sig_high - sig_low;

}


//----------------------------------------------------------------------------
//	apply_actfunc() 
//	  - apply activation function to all neurons
//----------------------------------------------------------------------------
void netlayer::apply_actfunc() {

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	output[i][j] = act_func(net_input[i][j]);
     }
  } 

}

//----------------------------------------------------------------------------
//	unuse_actfunc() 
//	  - do not use act function -- just copy netinput to output
//----------------------------------------------------------------------------
void netlayer::unuse_actfunc() {

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	output[i][j] = net_input[i][j];
     }
  } 

}

//----------------------------------------------------------------------------
//	get_aff_matrix() 
//	  - return neuron (i,j)'s aff weight matrix (from input layer k)
//----------------------------------------------------------------------------
matrix_coord netlayer::get_affw_matrix(int i, int j, int cur_inp) {

  matrix_coord mc;
  REAL mag_x, mag_y;
  int  cent_x, cent_y;

  // 00Nov5 if retina is padded, adjust mag_x and mag_y and the center
  // on the retinal space accordingly.
  if (padded_flag == 1 && cur_inp == 0) {
	mag_x = (REAL)cols/(REAL)(feedings[cur_inp]->cols-2*aff_rads[cur_inp]);
	mag_y = (REAL)rows/(REAL)(feedings[cur_inp]->rows-2*aff_rads[cur_inp]);

	cent_x = (int)(i/mag_x)+aff_rads[cur_inp];
	cent_y = (int)(j/mag_y)+aff_rads[cur_inp];
  // if it is not a padded retina, just use a straight ratio
  } else {
	mag_x = (REAL)cols/(REAL)feedings[cur_inp]->cols;
	mag_y = (REAL)rows/(REAL)feedings[cur_inp]->rows;

	cent_x = (int)(i/mag_x);
	cent_y = (int)(j/mag_y);
  }

  mc.matrix = wgt[i][j].aff_ws[cur_inp];
  mc.cx     = cent_x - rf_ranges[cur_inp][i][j].lowx;
  mc.cy     = cent_y - rf_ranges[cur_inp][i][j].lowy;
  mc.lowx   = rf_ranges[cur_inp][i][j].lowx;
  mc.lowy   = rf_ranges[cur_inp][i][j].lowy;
  mc.highx  = rf_ranges[cur_inp][i][j].highx;
  mc.highy  = rf_ranges[cur_inp][i][j].highy;
  mc.cols   = rf_ranges[cur_inp][i][j].cols;
  mc.rows   = rf_ranges[cur_inp][i][j].rows;

  return mc;
}

//----------------------------------------------------------------------------
//	get_inh_matrix() 
//	  - return neuron (i,j)'s inh weight matrix
//----------------------------------------------------------------------------
matrix_coord netlayer::get_inhw_matrix(int i, int j) {

  matrix_coord mc;

  int cent_x = i;
  int cent_y = j;

  mc.matrix = wgt[i][j].lat_inh_w;
  mc.cx     = cent_x - lat_inh_range[i][j].lowx;
  mc.cy     = cent_y - lat_inh_range[i][j].lowy;
  mc.lowx   = lat_inh_range[i][j].lowx;
  mc.lowy   = lat_inh_range[i][j].lowy;
  mc.highx  = lat_inh_range[i][j].highx;
  mc.highy  = lat_inh_range[i][j].highy;
  mc.cols   = lat_inh_range[i][j].cols;
  mc.rows   = lat_inh_range[i][j].rows;

  return mc;

}
  
//----------------------------------------------------------------------------
//	get_exc() 
//	  - return neuron (i,j)'s exc weight matrix
//----------------------------------------------------------------------------
matrix_coord netlayer::get_excw_matrix(int i, int j) {

  matrix_coord mc;

  int cent_x = i;
  int cent_y = j;

  mc.matrix = wgt[i][j].lat_exc_w;
  mc.cx     = cent_x - lat_exc_range[i][j].lowx;
  mc.cy     = cent_y - lat_exc_range[i][j].lowy;
  mc.lowx   = lat_exc_range[i][j].lowx;
  mc.lowy   = lat_exc_range[i][j].lowy;
  mc.highx  = lat_exc_range[i][j].highx;
  mc.highy  = lat_exc_range[i][j].highy;
  mc.cols   = lat_exc_range[i][j].cols;
  mc.rows   = lat_exc_range[i][j].rows;

  return mc;

}

//----------------------------------------------------------------------------
//	setup_weight() 
//	  - setup gaussian or other weights
//----------------------------------------------------------------------------
// square 
void netlayer::setup_weight(NETWGT_TYPE type, int radius, 
	REAL randomness, int cur) {

  matrix_coord matrix_info;

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
        switch (type) {
	  case NETWGT_CIRCULAR_AFF: // fall through
          case NETWGT_AFF   : matrix_info = get_affw_matrix(i,j,cur); break;
          case NETWGT_EXC   : matrix_info = get_excw_matrix(i,j); break;
          case NETWGT_INH   : matrix_info = get_inhw_matrix(i,j); break;
        }
        int cx = matrix_info.cx;
        int cy = matrix_info.cy;
        set_matrix(matrix_info.matrix,
		   matrix_info.cols,
		   matrix_info.rows,
		   cx,cy,radius,randomness);
	if (type == NETWGT_CIRCULAR_AFF)
		circular_affw(matrix_info.matrix,
			matrix_info.cols,matrix_info.rows,
			cx, cy, cur); 
     }
  }

}

//----------------------------------------------------------------------------
// gaussian blob
void netlayer::setup_weight(NETWGT_TYPE type, REAL sigma, REAL randomness,
	int cur) {

  matrix_coord matrix_info;

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
        switch (type) {
	  case NETWGT_CIRCULAR_AFF : // fall through
          case NETWGT_AFF   : matrix_info = get_affw_matrix(i,j,cur); break;
          case NETWGT_EXC   : matrix_info = get_excw_matrix(i,j); break;
          case NETWGT_INH   : matrix_info = get_inhw_matrix(i,j); break;
        }
        int cx = matrix_info.cx;
        int cy = matrix_info.cy;
        set_matrix(matrix_info.matrix,
		   matrix_info.cols,
		   matrix_info.rows,
		   cx,cy,sigma,randomness);
	if (type == NETWGT_CIRCULAR_AFF)
		circular_affw(matrix_info.matrix,
			matrix_info.cols,matrix_info.rows,
			cx, cy, cur); 
     }
  }

}

//----------------------------------------------------------------------------
// gaussian elongated blob
void netlayer::setup_weight(NETWGT_TYPE type, 
                REAL sigmax, REAL sigmay, REAL angle, 
		REAL randomness, int cur) {

  matrix_coord matrix_info;

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
        switch (type) {
	  case NETWGT_CIRCULAR_AFF : // fall through
          case NETWGT_AFF   : matrix_info = get_affw_matrix(i,j,cur); break;
          case NETWGT_EXC   : matrix_info = get_excw_matrix(i,j); break;
          case NETWGT_INH   : matrix_info = get_inhw_matrix(i,j); break;
        }
        int cx = matrix_info.cx;
        int cy = matrix_info.cy;
        set_matrix(matrix_info.matrix,
		   matrix_info.cols,
		   matrix_info.rows,
		   cx,cy,sigmax,sigmay,angle,randomness);
	if (type == NETWGT_CIRCULAR_AFF)
		circular_affw(matrix_info.matrix,
			matrix_info.cols,matrix_info.rows,
			cx, cy, cur); 
     }
  }

}

//----------------------------------------------------------------------------
//	set_matrix() 
//	  - set matrix values to a certain distribution
//	  - adopted from geninput.cc
//----------------------------------------------------------------------------
// square 
void  netlayer::set_matrix(REAL **matrix, int cols, int rows,
		int cx, int cy, int radius, REAL randomness) {

  int lowx = MAX(0,cx-radius);
  int lowy = MAX(0,cy-radius);
  int highx = MIN(cols-1, cx+radius);
  int highy = MIN(rows-1, cy+radius);

  // set input
  for (int i=lowx; i<=highx; ++i) {
     for (int j=lowy; j<=highy; ++j) {
        matrix[i][j] = (1.0-randomness) + randomness*drand48();
     }
  }
}
//----------------------------------------------------------------------------
// gaussian blob
void  netlayer::set_matrix(REAL **matrix, int cols, int rows,
		int cx, int cy, REAL sigma, REAL randomness) {

  // reuse oriented gaussian routine.
  set_matrix(matrix,cols,rows,cx,cy,sigma,sigma,0.0,randomness);

}

//----------------------------------------------------------------------------
// elongated gaussian blob
void  netlayer::set_matrix(REAL **matrix, int cols, int rows,
		int cx, int cy, 
		REAL sigmax, REAL sigmay, REAL angle, REAL randomness) {

  REAL radian = angle * M_PI / 180.0;
  REAL ci, cj, xsq, ysq;

  int radius = (int)(sqrt(sigmax*sigmax+sigmay*sigmay)+0.5)+2;
  int lowx = MAX(0,cx-radius);
  int lowy = MAX(0,cy-radius);
  int highx = MIN(cols-1, cx+radius);
  int highy = MIN(rows-1, cy+radius);

  for (int i=lowx; i<=highx; ++i) {
     for (int j=lowy; j<=highy; ++j) {
        ci = cx-(REAL)i;
        cj = cy-(REAL)j;
        xsq = ci*cos(radian) + cj*sin(radian);
        ysq = -ci*sin(radian) + cj*cos(radian);
        xsq *= xsq;
        xsq /= (sigmax*sigmax);
        ysq *= ysq;
        ysq /= (sigmay*sigmay);

        matrix[i][j] += (1.0-randomness)*exp(-(xsq+ysq))+randomness*drand48();
     }
  }

}

// testing testing 
int netlayer::addition (int argc, char *argv[]) {

  if (argc!=2) return -1;

  int val = atoi(argv[1]) + atoi(argv[2]);
  cout << "addition() result = " << val << endl;

  return val;
}

//----------------------------------------------------------------------------
// totally random
void  netlayer::set_matrix(REAL **matrix, int cols, int rows,
		REAL randomness) {

  for (int i=0; i<cols; ++i) {
     for (int j=0 ; j<rows; ++j) {
        matrix[i][j] = randomness*drand48()*2.0-1.0;
     }
  }
}


//----------------------------------------------------------------------------
//	get_affw() 
//	  - return afferent weight at unit (i, j), for input layer 'cur_inp'
//----------------------------------------------------------------------------
void netlayer::get_affw(REAL **dest_matrix, int col, int row, int cur_inp) {

  if (dest_matrix==NULL) {
	warning("netlayer::get_affw",
		"destination matrix is NULL, allocate it first"); 
	return;
  }

  // 1. first clear the matrix
  reset_matrix(dest_matrix,feedings[cur_inp]->cols,feedings[cur_inp]->rows);

  // 2. now copy the afferent weigth into there

  int basex = rf_ranges[cur_inp][col][row].lowx;
  int basey = rf_ranges[cur_inp][col][row].lowy;

  for (int i=rf_ranges[cur_inp][col][row].lowx;
           i<=rf_ranges[cur_inp][col][row].highx; ++i) {
     for (int j=rf_ranges[cur_inp][col][row].lowy;
              j<=rf_ranges[cur_inp][col][row].highy; ++j) {
	 
	if (wgt[col][row].aff_ws[cur_inp][i-basex][j-basey]>0.0) {
		dest_matrix[i][j] 
			= wgt[col][row].aff_ws[cur_inp][i-basex][j-basey];
	} else {	
		dest_matrix[i][j] = 0.0;
	}

     }
  }

}

//----------------------------------------------------------------------------
//	get_affw_only() 
//	  - return afferent weight at unit (i, j), for input layer 'cur_inp'
//	  - return only the RF 
//----------------------------------------------------------------------------
void netlayer::get_affw_only(REAL **dest_matrix, int col, int row, int cur_inp) {

  if (dest_matrix==NULL) {
	warning("netlayer::get_affw",
		"destination matrix is NULL, allocate it first"); 
	return;
  }

  // 1. first clear the matrix
  reset_matrix(dest_matrix,
		rf_ranges[cur_inp][col][row].cols,
		rf_ranges[cur_inp][col][row].rows);

  // 2. now copy the afferent weigth into there

  for (int i=0; i<rf_ranges[cur_inp][col][row].cols; ++i) {
     for (int j=0; j<rf_ranges[cur_inp][col][row].rows; ++j) {
	 
	if (wgt[col][row].aff_ws[cur_inp][i][j]>0.0) {
		dest_matrix[i][j] 
			= wgt[col][row].aff_ws[cur_inp][i][j];
	} else {	
		dest_matrix[i][j] = 0.0;
	}

     }
  }

}

//----------------------------------------------------------------------------
//	circular_affw() 
//	  - make RF circular by killing out of circle connections
//----------------------------------------------------------------------------
void netlayer::circular_affw(REAL **dest_matrix, 
		int cols, int rows, int cx, int cy, int cur_inp) {

  // kill out-of-circle connections by marking them with a small negative
  // value

  int radius = aff_rads[cur_inp];

  float radius_sq = (radius+0.5)*(radius+0.5);
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	if ((i-cx)*(i-cx)+(j-cy)*(j-cy) > radius_sq) {
		dest_matrix[i][j] = -DEATH_FLAG;
	}
     }
  }
  
}

//----------------------------------------------------------------------------
//	get_lat_excw() 
//	  - return lat_exc weight at unit (i, j)
//----------------------------------------------------------------------------
void netlayer::get_lat_excw(REAL **dest_matrix, int col, int row,
	int is_instar) {

  if (dest_matrix==NULL) {
	warning("netlayer::get_lat_excw",
		"destination matrix is NULL, allocate it first"); 
	return;
  }

  // 1. first clear the matrix
  reset_matrix(dest_matrix,cols,rows);

  // 2. now copy the afferent weigth into there


  if (is_instar) { // get weights when instar normalization is used

     int basei = lat_exc_range[col][row].lowx;
     int basej = lat_exc_range[col][row].lowy;

     for (int i=0; i<lat_exc_range[col][row].cols; ++i) {
        for (int j=0; j<lat_exc_range[col][row].rows; ++j) {

          if (wgt[col][row].lat_exc_w[i][j]>0.0)
	        dest_matrix[i+basei][j+basej] = wgt[col][row].lat_exc_w[i][j];
	  else
		dest_matrix[i+basei][j+basej] = KILLED;
	}
     }

  } else { // get weights when outstar normalization is used

     // x. dest_matrix[i][j] = wgt[col][row].lat_exc_w[i-basex][j-basey];
     // Remember that the weights are outstar, but the actual effect
     // is instar, so for a given neuron (col,row), the incoming 
     // weights wgt[i][j].[col][row] must be returned.
     for (int i=lat_exc_range[col][row].lowx;
           i<=lat_exc_range[col][row].highx; ++i) {
        for (int j=lat_exc_range[col][row].lowy;
              j<=lat_exc_range[col][row].highy; ++j) {
	 
	  if (wgt[i][j].lat_exc_w [col-lat_exc_range[i][j].lowx]
				  [row-lat_exc_range[i][j].lowy] > 0.0) 
	    	dest_matrix[i][j] = wgt[i][j].lat_exc_w
					[col-lat_exc_range[i][j].lowx]
					[row-lat_exc_range[i][j].lowy];
	  else 
	    dest_matrix[i][j] = KILLED;

	}
     }
  }

}

//----------------------------------------------------------------------------
//	get_lat_excw_only() 
//	  - return lat_exc weight at unit (i, j)
//	  - return area immediately reached by connections.
//----------------------------------------------------------------------------
void netlayer::get_lat_excw_only(REAL **dest_matrix, int col, int row, 
	int x_size, int y_size, 
	int is_instar) {

  if (dest_matrix==NULL) {
	warning("netlayer::get_lat_excw",
		"destination matrix is NULL, allocate it first"); 
	return;
  }

  // 1. first clear the matrix
  reset_matrix(dest_matrix,x_size,y_size);

  // 2. now copy the afferent weigth into there


  if (is_instar) { // get weights when instar normalization is used

     int basei = lat_exc_range[col][row].lowx;
     int basej = lat_exc_range[col][row].lowy;

     for (int i=0; i<lat_exc_range[col][row].cols; ++i) {
        for (int j=0; j<lat_exc_range[col][row].rows; ++j) {

          if (wgt[col][row].lat_exc_w[i][j]>0.0)
	        dest_matrix[i+basei][j+basej] = wgt[col][row].lat_exc_w[i][j];
	  else
		dest_matrix[i+basei][j+basej] = KILLED;
	}
     }

  } else { // get weights when outstar normalization is used

     // x. dest_matrix[i][j] = wgt[col][row].lat_exc_w[i-basex][j-basey];
     // Remember that the weights are outstar, but the actual effect
     // is instar, so for a given neuron (col,row), the incoming 
     // weights wgt[i][j].[col][row] must be returned.
     int basex = lat_exc_range[col][row].lowx;
     int basey = lat_exc_range[col][row].lowy;
     for (int i=basex; i<=lat_exc_range[col][row].highx; ++i) {
        for (int j=basey; j<=lat_exc_range[col][row].highy; ++j) {
	 
	  if (wgt[i][j].lat_exc_w [col-lat_exc_range[i][j].lowx]
				  [row-lat_exc_range[i][j].lowy] > 0.0) 
	    	dest_matrix[i-basex][j-basey] = wgt[i][j].lat_exc_w
					[col-lat_exc_range[i][j].lowx]
					[row-lat_exc_range[i][j].lowy];
	  else 
	    dest_matrix[i-basex][j-basey] = KILLED;

	}
     }
  }

}
//----------------------------------------------------------------------------
//	get_lat_inhw() 
//	  - return lat_inh weight at unit (i, j)
//----------------------------------------------------------------------------
void netlayer::get_lat_inhw(REAL **dest_matrix, int col, int row,
	int is_instar) {

  if (dest_matrix==NULL) {
	warning("netlayer::get_lat_inhw",
		"destination matrix is NULL, allocate it first"); 
	return;
  }

  // 1. first clear the matrix
  reset_matrix(dest_matrix,cols,rows);

  // 2. now copy the inhibitory weigth into there

  if (is_instar) { // get weights when instar normalization is used

     int basei = lat_inh_range[col][row].lowx;
     int basej = lat_inh_range[col][row].lowy;

     for (int i=0; i<lat_inh_range[col][row].cols; ++i) {
        for (int j=0; j<lat_inh_range[col][row].rows; ++j) {

          if (wgt[col][row].lat_inh_w[i][j]>0.0)
	        dest_matrix[i+basei][j+basej] = wgt[col][row].lat_inh_w[i][j];
	  else
		dest_matrix[i+basei][j+basej] = KILLED;
	}
     }

  } else { // get weights when outstar normalization is used

     // x. dest_matrix[i][j] = wgt[col][row].lat_inh_w[i-basex][j-basey];
     // Remember that the weights are outstar, but the actual effect
     // is instar, so for a given neuron (col,row), the incoming 
     // weights wgt[i][j].[col][row] must be returned.
     for (int i=lat_inh_range[col][row].lowx;
           i<=lat_inh_range[col][row].highx; ++i) {
        for (int j=lat_inh_range[col][row].lowy;
              j<=lat_inh_range[col][row].highy; ++j) {
	 
	  if (wgt[i][j].lat_inh_w [col-lat_inh_range[i][j].lowx]
				  [row-lat_inh_range[i][j].lowy] > 0.0) 
	    	dest_matrix[i][j] = wgt[i][j].lat_inh_w
					[col-lat_inh_range[i][j].lowx]
					[row-lat_inh_range[i][j].lowy];
	  else 
	    dest_matrix[i][j] = KILLED;

	}

     }
  }

}

//----------------------------------------------------------------------------
//	get_lat_inhw_only() 
//	  - return lat_inh weight at unit (i, j)
//	  - return only the area with actual connections
//----------------------------------------------------------------------------
void netlayer::get_lat_inhw_only(REAL **dest_matrix, int col, int row,
	int x_size, int y_size, 
	int is_instar) {

  if (dest_matrix==NULL) {
	warning("netlayer::get_lat_inhw",
		"destination matrix is NULL, allocate it first"); 
	return;
  }

  // 1. first clear the matrix
  reset_matrix(dest_matrix,x_size,y_size);

  // 2. now copy the inhibitory weigth into there

  if (is_instar) { // get weights when instar normalization is used

     int basei = lat_inh_range[col][row].lowx;
     int basej = lat_inh_range[col][row].lowy;

     for (int i=0; i<lat_inh_range[col][row].cols; ++i) {
        for (int j=0; j<lat_inh_range[col][row].rows; ++j) {

          if (wgt[col][row].lat_inh_w[i][j]>0.0)
	        dest_matrix[i+basei][j+basej] = wgt[col][row].lat_inh_w[i][j];
	  else
		dest_matrix[i+basei][j+basej] = KILLED;
	}
     }

  } else { // get weights when outstar normalization is used

     // x. dest_matrix[i][j] = wgt[col][row].lat_inh_w[i-basex][j-basey];
     // Remember that the weights are outstar, but the actual effect
     // is instar, so for a given neuron (col,row), the incoming 
     // weights wgt[i][j].[col][row] must be returned.
     int basex = lat_inh_range[col][row].lowx;
     int basey = lat_inh_range[col][row].lowy;
     for (int i=basex; i<=lat_inh_range[col][row].highx; ++i) {
        for (int j=basey; j<=lat_inh_range[col][row].highy; ++j) {
	 
	  if (wgt[i][j].lat_inh_w [col-lat_inh_range[i][j].lowx]
				  [row-lat_inh_range[i][j].lowy] > 0.0) 
	    	dest_matrix[i-basex][j-basey] = wgt[i][j].lat_inh_w
					[col-lat_inh_range[i][j].lowx]
					[row-lat_inh_range[i][j].lowy];
	  else 
	    dest_matrix[i-basex][j-basey] = KILLED;

	}

     }
  }

}
//----------------------------------------------------------------------------
//	get_lat_combinedw() 
//	  - return lat_combined weight at unit (i, j)
//----------------------------------------------------------------------------
void netlayer::get_lat_combinedw(REAL **dest_matrix, int col, int row) {

  if (dest_matrix==NULL) {
	warning("netlayer::get_lat_excw",
		"destination matrix is NULL, allocate it first"); 
	return;
  }

  // 1. first clear the matrix
  reset_matrix(dest_matrix,cols,rows);

  // 2. now copy the excitatory weigth into there

  for (int i=lat_exc_range[col][row].lowx;
           i<=lat_exc_range[col][row].highx; ++i) {
     for (int j=lat_exc_range[col][row].lowy;
              j<=lat_exc_range[col][row].highy; ++j) {
	 
	// Remember that the weights are outstar, but the actual effect
	// is instar, so for a given neuron (col,row), the incoming 
	// weights wgt[i][j].[col][row] must be returned.
	dest_matrix[i][j] = wgt[i][j].lat_exc_w
				[col-lat_exc_range[i][j].lowx]
				[row-lat_exc_range[i][j].lowy];

     }
  }

  // 3. now copy the inhibitory weigth into there

  for (int i=lat_inh_range[col][row].lowx;
           i<=lat_inh_range[col][row].highx; ++i) {
     for (int j=lat_inh_range[col][row].lowy;
              j<=lat_inh_range[col][row].highy; ++j) {
	 
	// Remember that the weights are outstar, but the actual effect
	// is instar, so for a given neuron (col,row), the incoming 
	// weights wgt[i][j].[col][row] must be returned.
	dest_matrix[i][j] -= wgt[i][j].lat_inh_w
				[col-lat_inh_range[i][j].lowx]
				[row-lat_inh_range[i][j].lowy];

     }
  }

  // 4. now scale it.
  for (int i=0; i<cols; ++i){
     for (int j=0; j<rows; ++j) {
	 
	dest_matrix[i][j] = (dest_matrix[i][j]-(-0.07))/(1.0-(-0.07));

     }
  }

}


//----------------------------------------------------------------------------
//	kill_overlapping_inh_w() 
//	  - kill central rectangular region in all inhibitory weights
//	    that overlaps with the excitatory weight.
//----------------------------------------------------------------------------
void netlayer::kill_overlapping_inh_w() {

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	int basex = lat_inh_range[i][j].lowx;
	int basey = lat_inh_range[i][j].lowy;

	for (int x=lat_exc_range[i][j].lowx; 
		 x<=lat_exc_range[i][j].highx; ++x)
	   for (int y=lat_exc_range[i][j].lowy;
		    y<=lat_exc_range[i][j].highy;++y) 
		
		wgt[i][j].lat_inh_w[x-basex][y-basey] *= -1;

     }
  }

}

//----------------------------------------------------------------------------
//	calc_or_pref() 
//	  - calculate orientation pref/selectivity
//	    and save in or_pref[][].
//	  - degrees are encoded within [0.0-1.0] == [0deg - 180deg]
//	  - sigmax and sigmay represents the gaussian blob's size
//----------------------------------------------------------------------------
void netlayer::calc_or_pref(int cur_inp, REAL sigmax, REAL sigmay) {

  // calculate COG for later use
  calc_cog(cur_inp);

  // clear the array
  for (int i=0; i<cols; ++i)
      for (int j=0; j<rows; ++j)
	  or_pref[i][j] = 0.0;

  // for all units
  for (int i=0; i<cols; ++i) {
      cerr << "calc_or_pref : col" << i << endl;
      for (int j=0; j<rows; ++j) 
	   calc_single_or_pref_vecsum(cur_inp,i,j, sigmax, sigmay, 
		&(or_pref[i][j]), &(or_selectivity[i][j]));
  }

}

//----------------------------------------------------------------------------
//	calc_single_or_pref() 
//	  -  calculate orientation pref/selectivity for a single neuron
//----------------------------------------------------------------------------
REAL netlayer::calc_single_or_pref(int cur_inp, int x, int y, 
	REAL sigmax, REAL sigmay) {

  // for 0 to 180 degrees, place gaussian blob on the COG of that
  // neuron's afferent weight matrix, and calculate the maximally
  // responding one, and return it in 0.0(0deg)-1.0(180deg) scale.
  int samples = 18;
  int x_size  = rf_ranges[cur_inp][x][y].cols;
  int y_size  = rf_ranges[cur_inp][x][y].rows;

  register REAL tmp_wgt;
  register REAL **tmp_wgt_pp;

  REAL mult = 180.0/(REAL)samples;

  REAL max_resp = -99999.9;
  int  max_idx  = 0;
  for (int k=0; k<samples; ++k) { // for 0 - 180 degree rotation

    for (int p=0; p<x_size; ++p) { // for a complete sweep from left to right
      // generate gaussian
      for (int i=0; i<x_size; ++i)
          for (int j=0; j<y_size; ++j) 
		ref_pattern[i][j] = 0.0;

      if (k<samples/2) {
      		geninput::gaussian(ref_pattern, x_size, y_size,
			x_size-p,p, 
			(REAL)(x_size*2.0), 1.0, k*mult);
      } else {
      		geninput::gaussian(ref_pattern, x_size, y_size,
			x_size-p, y_size-p,
			(REAL)(x_size*2.0), 1.0, k*mult);
      } 

      /*
      geninput::gaussian(ref_pattern, x_size, y_size,
		(int)cogs[cur_inp][x][y].x-rf_ranges[cur_inp][x][y].lowx, 
		(int)cogs[cur_inp][x][y].y-rf_ranges[cur_inp][x][y].lowy,	
		sigmax, sigmay, (REAL)k * 12.0);
      geninput::sine_grating(ref_pattern,x_size, y_size, DONTCARE, DONTCARE,
		x_size/2.0,0.0,(REAL)k * 12.0);
      */

      // calc dotproduct (as a vector)
      REAL dot = 0.0;
      // get pointer to afferent weight matrix for [x][y] to save on
      // indirect memory reference overhead.
      tmp_wgt_pp = wgt[x][y].aff_ws[cur_inp];
      for (int i=0; i<x_size; ++i)
          for (int j=0; j<y_size; ++j) {
		tmp_wgt = tmp_wgt_pp[i][j];
	      	if (tmp_wgt>0.0) {
	          dot += ref_pattern[i][j]*tmp_wgt;
	      	}
	   }
      //debug: cout << dot;
  
      if (dot>=max_resp) {
	  max_resp = dot;
	  max_idx  = k;
	  //debug: cout << " " << k << " *" << endl;
      } else {
	  //debug: cout << endl;
      }
    }
    //debug: cout << "-------" << endl;

  }
 
  //debug: cout << "------------------" << endl;

  return (max_idx/(REAL)(samples));

} 

//----------------------------------------------------------------------------
//	calc_single_or_pref_vecsum() 
//	  - calculate or pref : copied from netlayer.cc
//	  - use jbednar's vector sum method 
//----------------------------------------------------------------------------
void netlayer::calc_single_or_pref_vecsum(int cur_inp, int x, int y, 
	REAL sigmax, REAL sigmay, REAL *orpref, REAL *orselec) {

  int samples = 18;
  int step = samples/6;
  int x_size  = rf_ranges[cur_inp][x][y].cols;
  int y_size  = rf_ranges[cur_inp][x][y].rows;

  REAL max_or_resp[samples];

  register REAL tmp_wgt;
  register REAL **tmp_wgt_pp;

  REAL mult = 180.0/(REAL)samples;

  REAL max_resp = -99999.9;
  int  max_idx  = 0;

  // 0. initialize max_or_resp array;
  for (int i=0; i<samples; i+=step) {
        max_or_resp[i] = 0.0;
  }

  // 1. Big loop : for 0 to 180 degree rotation
  for (int k=0; k<samples; k += step) { // for 0 - 180 degree rotation

    // 1.1. Small loop for a complete sweep from one side to another
    for (int p=0; p<MAX(x_size,y_size); ++p) { 

      // 1.1.1 generate gaussian
      for (int i=0; i<x_size; ++i)
          for (int j=0; j<y_size; ++j)
                ref_pattern[i][j] = 0.0;

      if (k<samples/2) {
                geninput::gaussian(ref_pattern, x_size, y_size,
                        (int)x_size-p,p, 
			(REAL)x_size*2.0, (REAL)1.0, (REAL)(k * mult));
      } else {
                geninput::gaussian(ref_pattern, x_size, y_size,
                        x_size-p, (int)y_size-p,
			(REAL)x_size*2.0, (REAL)1.0, (REAL)(k * mult));
      }

      // 1.1.2 calc dotproduct (as a vector)
      REAL dot = 0.0;
      tmp_wgt_pp = wgt[x][y].aff_ws[cur_inp];
      for (int i=0; i<x_size; ++i) {
          for (int j=0; j<y_size; ++j) {
		tmp_wgt = tmp_wgt_pp[i][j];
	      	if (tmp_wgt>0.0) {
	          dot += ref_pattern[i][j]*tmp_wgt;
	      	}
	   }
      }

      // 1.1.3 update max_or_resp for 
      if (dot>=max_or_resp[k]) {
          max_or_resp[k] = max_resp = dot;
	  max_idx  = k;
      } 
    }
  }

  // 2. Calculate vector sum
  REAL vector_sum = 0.0;
  REAL cart_x = 0.0;
  REAL cart_y = 0.0;  
  REAL selectivity = 0.0;

  mult = 2.0*10.0*3.141592/180.0;

  for (int i=0; i<samples; i+=step) {
	cart_x += max_or_resp[i] * cos(i*mult);
	cart_y += max_or_resp[i] * sin(i*mult);
  }

  vector_sum = 180.0*atan2(cart_y,cart_x)/3.141592/2.0; 
  

  if (vector_sum < 0.0) vector_sum += 180.0;

  *orpref = (vector_sum/180.0); // 0.0 == 0 degree, 1.0 == 180 degrees
  *orselec = sqrt(cart_y*cart_y+cart_x*cart_x);

}

//----------------------------------------------------------------------------
//	print_or_pref() 
//	  - print out the orientation preference
//----------------------------------------------------------------------------
void netlayer::print_or_pref() {

  for (int j=0; j<rows; ++j) {
      for (int i=0; i<cols; ++i) {
	   printf("%.3f ",or_pref[i][j]);
      }
      cout << endl;
  }

}
//----------------------------------------------------------------------------
void netlayer::save_or_pref(int cur_inp, char *filename) {

  // Must run calc_or_pref first !  

  FILE *fp = fopen(filename,"w");

  if (fp==NULL) { 
	warning("netlayer::save_or_pref","file error : %s",filename);
	return;
  } 

  // calc_or_pref(cur_inp,cols,1.0);

  for (int j=0; j<rows; ++j) {
      for (int i=0; i<cols; ++i) {
           fprintf(fp,"%.3f ",or_pref[i][j]);
      }
      fprintf(fp,"\n");
  }

  fclose(fp);

}
//----------------------------------------------------------------------------
void netlayer::save_or_selectivity(int cur_inp, char *filename) {

  // Must run calc_or_pref first !  

  FILE *fp = fopen(filename,"w");

  if (fp==NULL) { 
	warning("netlayer::save_or_selectivity","file error : %s",filename);
	return;
  } 

  // calc_or_pref(cur_inp,cols,1.0);

  for (int j=0; j<rows; ++j) {
      for (int i=0; i<cols; ++i) {
           fprintf(fp,"%.3f ",or_selectivity[i][j]);
      }
      fprintf(fp,"\n");
  }

  fclose(fp);

}
//----------------------------------------------------------------------------
int netlayer::call_save_or_pref(int argc, char *argv[]) {

  // argv[1] = cur_input, argv[2] = filename

  if (argc!=2) 
	return 0;

  save_or_pref(atoi(argv[1]),argv[2]);

  return 1;

}
//----------------------------------------------------------------------------
int netlayer::call_save_or_selectivity(int argc, char *argv[]) {

  // argv[1] = cur_input, argv[2] = filename

  if (argc!=2) 
	return 0;

  save_or_selectivity(atoi(argv[1]),argv[2]);

  return 1;

}

//----------------------------------------------------------------------------
//      calc_cog() 
//        - calculate center of gravity coordinates
//----------------------------------------------------------------------------
void netlayer::calc_cog(int cur_inp_layer){

  REAL sum, x_sum, y_sum;

  for (int x=0; x<cols; ++x) {

     for (int y=0; y<rows; ++y) {

        sum = 0.0;
        x_sum = 0.0;
        y_sum = 0.0;
        for (int i=rf_ranges[cur_inp_layer][x][y].lowx;
                 i<=rf_ranges[cur_inp_layer][x][y].highx; ++i) {
           for (int j=rf_ranges[cur_inp_layer][x][y].lowy;
                    j<=rf_ranges[cur_inp_layer][x][y].highy; ++j) {
                int rf_x = i-rf_ranges[cur_inp_layer][x][y].lowx;
                int rf_y = j-rf_ranges[cur_inp_layer][x][y].lowy;
                REAL w = wgt[x][y].aff_ws[cur_inp_layer][rf_x][rf_y];
                sum += w;
                x_sum += i*w;
                y_sum += j*w;
           }
        }

        cogs[cur_inp_layer][x][y].x = x_sum/sum;
        cogs[cur_inp_layer][x][y].y = y_sum/sum;

     }
  }
}


//----------------------------------------------------------------------------
//	dump_wgts() 
//	  - dump afferent and lateral weights of a given unit
//----------------------------------------------------------------------------
void netlayer::dump_wgts(int col, int row, int cur_inp, 
		NETWGT_TYPE choice, FILE *fp) {

  // afferent weight matrix dimension
  int  aff_x = feedings[cur_inp]->cols;
  int  aff_y = feedings[cur_inp]->rows;
  matrix<REAL> *buffer;

  switch (choice) {

  case NETWGT_AFF: 
	  // 1. afferent weights
	  // cout << "# affw : " << col << " " << row << endl;
	  // cout << aff_x << " " << aff_y << endl;
	  fprintf(fp,"# affw : %d %d\n",col,row);
	  fprintf(fp,"%d %d\n",aff_x,aff_y);
	  buffer = new matrix<REAL>(aff_x,aff_y);
	  get_affw(buffer->mat,col,row,cur_inp);
	  buffer->print_matrix(fp);
	  delete buffer;
	  break;

  case NETWGT_EXC:
	  // alloc buffer 
	  buffer = new matrix<REAL>(cols,rows);

	  // 2. lat exc weights
	  // cout << "# excw : " << col << " " << row << endl;
	  // cout << cols << " " << rows << endl;
	  fprintf(fp,"# excw : %d %d\n",col,row);
	  fprintf(fp,"%d %d\n",cols,rows);
	  get_lat_excw(buffer->mat,col,row,cur_inp);
	  buffer->print_matrix(fp);
	  delete buffer;
	  break;
	  
  case NETWGT_INH:
	  // alloc buffer 
	  buffer = new matrix<REAL>(cols,rows);

	  // 3. lat inh weights
	  // cout << "# inhw : " << col << " " << row << endl;
	  // cout << cols << " " << rows << endl;
	  fprintf(fp,"# inhw : %d %d\n",col,row);
	  fprintf(fp,"%d %d\n",cols,rows);
	  get_lat_inhw(buffer->mat,col,row,cur_inp);
	  buffer->print_matrix(fp);
	  delete buffer;
	  break;

   case NETWGT_CIRCULAR_AFF: break;

  }
}
//----------------------------------------------------------------------------
int netlayer::call_dump_wgts(int argc, char *argv[]) {

  int cur_inp;

  if (argc==2) 
	cur_inp = 0;
  else if (argc==3) 
	cur_inp = atoi(argv[3]);
  else
	return 0;

  FILE *fp;

  cerr << "netlayer::call_dump_wgts - appending to dump.w\n";
  fp = fopen("dump.w","a");
  if (!fp) {
	fp = stdout;
  }
  dump_wgts_only(atoi(argv[1]),atoi(argv[2]),cur_inp,NETWGT_AFF,fp);
  dump_wgts_only(atoi(argv[1]),atoi(argv[2]),cur_inp,NETWGT_EXC,fp);
  dump_wgts_only(atoi(argv[1]),atoi(argv[2]),cur_inp,NETWGT_INH,fp);
  fclose(fp);

  return 1;
}

//----------------------------------------------------------------------------
int netlayer::call_dump_wgts_range(int argc, char *argv[]) {

  int cur_inp;

  if (argc==4) 
	cur_inp = 0;
  else if (argc==5) 
	cur_inp = atoi(argv[5]);
  else 
	return 0;

  int x1 = atoi(argv[1]); int y1 = atoi(argv[2]);
  int x2 = atoi(argv[3]); int y2 = atoi(argv[4]);

  int xstart = MIN(x1,x2); int ystart = MIN(y1,y2);
  int xend   = MAX(x1,x2); int yend   = MAX(y1,y2);

  // 00sep23
  if (xstart<0) xstart=0; 
  if (xstart>=cols) xstart=cols-1; 
  if (ystart<0) ystart=0; 
  if (ystart>=rows) ystart=rows-1; 

  cerr << "netlayer::call_dump_wgts_range - writing to dump.w\n";
  FILE *fp;
  fp = fopen("dump.w","w");
  if (!fp) {
	fp = stdout;
  }
  for (int i=xstart; i<=xend; ++i) {
     for (int j=ystart; j<=yend; ++j) {
  	dump_wgts_only(i,j,cur_inp,NETWGT_AFF,fp);
  	dump_wgts_only(i,j,cur_inp,NETWGT_EXC,fp);
  	dump_wgts_only(i,j,cur_inp,NETWGT_INH,fp);
     }
  }
  fclose(fp);

  return 1;

}

//----------------------------------------------------------------------------
int netlayer::call_dump_wgts_range_select(int argc, char *argv[]) {

  int cur_inp, select;

  if (argc==4) {
	cur_inp = 0;
	select = (int)NETWGT_AFF;
  } else if (argc==5) {
	cur_inp = atoi(argv[5]);
	select = (int)NETWGT_AFF;
  } else if (argc==6) {
	cur_inp = atoi(argv[5]);
	select = atoi(argv[6]);	
  } else 
	return 0;

  int x1 = atoi(argv[1]); int y1 = atoi(argv[2]);
  int x2 = atoi(argv[3]); int y2 = atoi(argv[4]);

  int xstart = MIN(x1,x2); int ystart = MIN(y1,y2);
  int xend   = MAX(x1,x2); int yend   = MAX(y1,y2);

  cerr << "netlayer::call_dump_wgts_range_select - writing to dump.w\n";
  FILE *fp;
  fp = fopen("dump.w","w");
  if (!fp) {
	fp = stdout;
  }
  for (int i=xstart; i<=xend; ++i) {
     for (int j=ystart; j<=yend; ++j) {
  	dump_wgts_only(i,j,cur_inp,(NETWGT_TYPE)select,fp);
     }
  }
  fclose(fp);

  return 1;

}

//----------------------------------------------------------------------------
//	dump_wgts_only() 
//	  - dump afferent and lateral weights of a given unit
//	  - only dump weights of area with actual connections
//----------------------------------------------------------------------------
void netlayer::dump_wgts_only(int col, int row, int cur_inp, 
		NETWGT_TYPE choice, FILE *fp) {

  // afferent weight matrix dimension
  int  aff_x = rf_ranges[cur_inp][col][row].cols;
  int  aff_y = rf_ranges[cur_inp][col][row].rows;
  // lat exc weight matrix dimension
  int  exc_x = lat_exc_range[col][row].cols;
  int  exc_y = lat_exc_range[col][row].rows;
  // lat inh weight matrix dimension
  int  inh_x = lat_inh_range[col][row].cols;
  int  inh_y = lat_inh_range[col][row].rows;

  matrix<REAL> *buffer;

  switch (choice) {

  case NETWGT_AFF: 
	  // 1. afferent weights
	  // cout << "# affw : " << col << " " << row << endl;
	  // cout << aff_x << " " << aff_y << endl;
	  fprintf(fp,"# affw : %d %d\n",col,row);
	  fprintf(fp,"%d %d\n",aff_x,aff_y);
	  buffer = new matrix<REAL>(aff_x,aff_y);
	  get_affw_only(buffer->mat,col,row,cur_inp);
	  buffer->print_matrix(fp);
	  delete buffer;
	  break;

  case NETWGT_EXC:
	  // alloc buffer 
	  buffer = new matrix<REAL>(exc_x,exc_y);

	  // 2. lat exc weights
	  // cout << "# excw : " << col << " " << row << endl;
	  // cout << exc_x << " " << exc_y << endl;
	  fprintf(fp,"# excw : %d %d\n",col,row);
	  fprintf(fp,"%d %d\n",exc_x,exc_y);
	  get_lat_excw_only(buffer->mat,col,row,exc_x,exc_y,cur_inp);
	  buffer->print_matrix(fp);
	  delete buffer;
	  break;
	  
  case NETWGT_INH:
	  // alloc buffer 
	  buffer = new matrix<REAL>(inh_x,inh_y);

	  // 3. lat inh weights
	  // cout << "# inhw : " << col << " " << row << endl;
	  // cout << inh_x << " " << inh_y << endl;
	  fprintf(fp,"# inhw : %d %d\n",col,row);
	  fprintf(fp,"%d %d\n",inh_x,inh_y);
	  get_lat_inhw_only(buffer->mat,col,row,inh_x,inh_y,cur_inp);
	  buffer->print_matrix(fp);
	  delete buffer;
	  break;

   case NETWGT_CIRCULAR_AFF: break;

  }
}

//----------------------------------------------------------------------------
//	calc_contrib_sum_mult() 
//	  - calculate net_input from aff and lateral contribs
//----------------------------------------------------------------------------
void netlayer::calc_contrib_sum_mult(int calc_aff){
  //
  // Calculate net_input from calculated aff(with calc_affresp()) and 
  // lat responses (calculated right above).
  //

  reset_matrix(net_input,cols,rows);

  if (calc_aff) {

     	for (int i=0; i<cols; ++i) {
     	   for (int j=0; j<rows; ++j) {
		net_input[i][j] = net_aff_input[i][j];
     	   }
  	}

  } else {

     	for (int i=0; i<cols; ++i) {
     	   for (int j=0; j<rows; ++j) {
		// net_aff_input[i][j] has already been multiplied
		// by aff_contris[]
		net_input[i][j] = net_aff_input[i][j] 
		   * (1.0 + lat_exc_contri * net_exc_input[i][j]) 
		   - lat_inh_contri * net_inh_input[i][j];
     	   }
  	}
  }

}

//----------------------------------------------------------------------------
//	dump_wgt_sums() 
//	  - dump afferent and lateral weights of a given unit
//----------------------------------------------------------------------------
void netlayer::dump_wgt_sums(int col, int row, int cur_inp) {

  // afferent weight matrix dimension
  int  aff_x = feedings[cur_inp]->cols;
  int  aff_y = feedings[cur_inp]->rows;
  REAL min,max;

  // 1. afferent weights
  cout << "# affw : " << endl;
  matrix<REAL> *buffer = new matrix<REAL>(aff_x,aff_y);
  get_affw(buffer->mat,col,row,cur_inp);
  buffer->minmax(&min,&max);
  cout << buffer->sum() << " " << min << " " << max << endl;
  delete buffer;

  // alloc buffer for use with exc and inh
  buffer = new matrix<REAL>(cols,rows);

  // 2. lat exc weights
  cout << "# excw : " << col << " " << row << endl;
  cout << cols << " " << rows << endl;
  get_lat_excw(buffer->mat,col,row,cur_inp);
  buffer->minmax(&min,&max);
  cout << buffer->sum() << " " << min << " " << max << endl;

  // 3. lat inh weights
  cout << "# inhw : " << col << " " << row << endl;
  cout << cols << " " << rows << endl;
  get_lat_inhw(buffer->mat,col,row,cur_inp);
  buffer->minmax(&min,&max);
  cout << buffer->sum() << " " << min << " " << max << endl;

  // clean up
  delete buffer;

}
//----------------------------------------------------------------------------
int netlayer::call_dump_wgt_sums(int argc, char *argv[]) {

  int cur_inp;

  if (argc==2) 
	cur_inp = 0;
  else if (argc==3) 
	cur_inp = atoi(argv[3]);
  else
	return 0;

  dump_wgt_sums(atoi(argv[1]),atoi(argv[2]),cur_inp);

  return 1;

}

//----------------------------------------------------------------------------
int netlayer::call_dump_wgt_sums_range(int argc, char *argv[]) {

  int cur_inp;

  if (argc==4) 
	cur_inp = 0;
  else if (argc==5) 
	cur_inp = atoi(argv[5]);
  else
	return 0;

  int x1 = atoi(argv[1]); int y1 = atoi(argv[2]);
  int x2 = atoi(argv[3]); int y2 = atoi(argv[4]);

  int xstart = MIN(x1,x2); int ystart = MIN(y1,y2);
  int xend   = MAX(x1,x2); int yend   = MAX(y1,y2);

  for (int i=xstart; i<=xend; ++i) {
     for (int j=ystart; j<=yend; ++j) {
	dump_wgt_sums(i,j,cur_inp);
     }
  }

  return 1;

}

//----------------------------------------------------------------------------
//      save_cog() 
//        - save center of gravity coordinates
//----------------------------------------------------------------------------
void netlayer::save_cog(int cur_inp_layer, char *filename){

  ofstream outfile(filename,ios::out); 

  if (outfile.bad()) {
	cerr << "netlayer::save_cog - " << filename << " cannot be opened\n";
	return;
  } else {
	cerr << "netlayer::save_cog - " << filename << " opened\n";
  }

  for (int y=0; y<rows; ++y) {

  	for (int x=0; x<cols; ++x) {
        	outfile << cogs[cur_inp_layer][x][y].x 
			<< ":"
        		<< cogs[cur_inp_layer][x][y].y << " ";
     	}
	outfile << endl;
	
  }

}

//----------------------------------------------------------------------------
//	dump_wgt_or_histo() 
//	  - dump the distribution of orientation preference
//	    of the lateral weights (with values above a certain threshold)
//----------------------------------------------------------------------------
void netlayer::dump_wgt_or_histo(int col, int row, REAL thresh,
		NETWGT_TYPE choice, FILE *fp, int omit_radius) {

  // PRE-conditions: 
  //	Assume or_pref contains the current or_preference of this map

  matrix<REAL> *buffer;
  // alloc buffer 
  buffer = new matrix<REAL>(cols,rows);

  int histo[35];

  switch (choice) {

  case NETWGT_EXC:
	  // 2. lat exc weights
	  // cout << "# excw : " << col << " " << row << endl;
	  // cout << cols << " " << rows << endl;
	  fprintf(fp,"# excw : %d %d\n",col,row);
	  get_lat_excw(buffer->mat,col,row,0);
	  break;
	  
  case NETWGT_INH:
	  // alloc buffer 
	  // 3. lat inh weights
	  // cout << "# inhw : " << col << " " << row << endl;
	  // cout << cols << " " << rows << endl;
	  fprintf(fp,"# inhw : %d %d\n",col,row);
	  get_lat_inhw(buffer->mat,col,row,0);
	  break;
  default: 
	  // not supported
	  cout << "netlayer::dump_wgt_or_histo : unsupported\n";
	  cerr << "netlayer::dump_wgt_or_histo : unsupported\n";
	  break;
  }

  // Now buffer contains the weight matrix.
  //
  // compare orientation of unit (col,row) and
  // other units connected with lateral connection
  // greater than "thresh" and increment count in
  // histogram.
  // The histogram is indexed by the difference in 
  // orientation. 

  // Remove omitted area from the center (col,row).
  // unimplemented.

  for (int i=0; i<35; ++i) {
	histo[i] = 0;
  }

  int cur_pref = (int)(or_pref[col][row]*180.0+0.5);
  int diff;
  for (int i=0; i<cols; ++i) {
    for (int j=0; j<rows; ++j) {
	if (buffer->mat[i][j] > thresh) {
		int target_pref = (int)(or_pref[i][j]*180.0+0.5);
		// diff = cur_pref-target_pref+170;
		if (target_pref<cur_pref) {
			diff = target_pref + 180 - cur_pref;
		} else {
			diff = target_pref - cur_pref;
		}
		histo[(int)(diff/10)]++;
	}
    }
  }

  // print result
  for (int i=0; i<18; ++i) {
	fprintf(fp,"%d ",histo[i]);
  }
  fprintf(fp,"\n");

  delete buffer;

}
//----------------------------------------------------------------------------
int netlayer::call_dump_wgt_or_histo(int argc, char *argv[]) {

  // call_dump_wgt_or_histo <startx> <starty> <endx> <endy> <thresh>

  if (argc!=5) {
	return 0;
  }

  int x1 = atoi(argv[1]); int y1 = atoi(argv[2]);
  int x2 = atoi(argv[3]); int y2 = atoi(argv[4]);
 
  REAL thresh = atof(argv[5]);

  int xstart = MIN(x1,x2); int ystart = MIN(y1,y2);
  int xend   = MAX(x1,x2); int yend   = MAX(y1,y2);

  FILE *fp = fopen("wgt_or.histo","w");

  if (!fp) {
	warning("call_dump_wgt_or_histo","cannot write to wgt_or.histo");
	return 0;
  }

  for (int i=xstart; i<=xend; ++i) {
     for (int j=ystart; j<=yend; ++j) {
	dump_wgt_or_histo(i,j,thresh,NETWGT_EXC,fp);
     }
  }
  for (int i=xstart; i<=xend; ++i) {
     for (int j=ystart; j<=yend; ++j) {
	dump_wgt_or_histo(i,j,thresh,NETWGT_INH,fp);
     }
  }

  fclose(fp);

  return 1;

}

//----------------------------------------------------------------------------
//	dump_wgt_or_histo_weighted() 
//	  - dump the distribution of orientation preference
//	    of the lateral weights (with values above a certain threshold)
//----------------------------------------------------------------------------
void netlayer::dump_wgt_or_histo_weighted(int col, int row, REAL thresh,
		NETWGT_TYPE choice, FILE *fp) {

  // PRE-conditions: 
  //	Assume or_pref contains the current or_preference of this map

  matrix<REAL> *buffer;
  // alloc buffer 
  buffer = new matrix<REAL>(cols,rows);

  REAL histo[18];

  switch (choice) {

  case NETWGT_EXC:
	  // 2. lat exc weights
	  // cout << "# excw : " << col << " " << row << endl;
	  // cout << cols << " " << rows << endl;
	  fprintf(fp,"# excw : %d %d\n",col,row);
	  get_lat_excw(buffer->mat,col,row,0);
	  break;
	  
  case NETWGT_INH:
	  // alloc buffer 
	  // 3. lat inh weights
	  // cout << "# inhw : " << col << " " << row << endl;
	  // cout << cols << " " << rows << endl;
	  fprintf(fp,"# inhw : %d %d\n",col,row);
	  get_lat_inhw(buffer->mat,col,row,0);
	  break;
  default: 
	  // not supported
	  cout << "netlayer::dump_wgt_or_histo : unsupported\n";
	  cerr << "netlayer::dump_wgt_or_histo : unsupported\n";
	  break;
  }

  // Now buffer contains the weight matrix.
  //
  // compare orientation of unit (col,row) and
  // other units connected with lateral connection
  // greater than "thresh" and increment count in
  // histogram.
  // The histogram is indexed by the difference in 
  // orientation. 

  for (int i=0; i<18; ++i) {
	histo[i] = 0.0;
  }

  int cur_pref = (int)(or_pref[col][row]*180.0+0.5);
  for (int i=0; i<cols; ++i) {
    for (int j=0; j<rows; ++j) {
	if (buffer->mat[i][j] > thresh) {
		int target_pref = (int)(or_pref[i][j]*180.0+0.5);
		int diff = MAX(cur_pref,target_pref)-MIN(cur_pref,target_pref);
		histo[(int)(diff/10)] += buffer->mat[i][j];
	}
    }
  }

  // print result
  for (int i=0; i<18; ++i) {
	fprintf(fp,"%f ",histo[i]);
  }
  fprintf(fp,"\n");

  delete buffer;

}
//----------------------------------------------------------------------------
int netlayer::call_dump_wgt_or_histo_weighted(int argc, char *argv[]) {

  // call_dump_wgt_or_histo <startx> <starty> <endx> <endy> <thresh>

  if (argc!=5) {
	return 0;
  }

  int x1 = atoi(argv[1]); int y1 = atoi(argv[2]);
  int x2 = atoi(argv[3]); int y2 = atoi(argv[4]);
 
  REAL thresh = atof(argv[5]);

  int xstart = MIN(x1,x2); int ystart = MIN(y1,y2);
  int xend   = MAX(x1,x2); int yend   = MAX(y1,y2);

  FILE *fp = fopen("wgt_or_weighted.histo","w");

  if (!fp) {
	warning("call_dump_wgt_or_histo","cannot write to wgt_or.histo");
	return 0;
  }

  for (int i=xstart; i<=xend; ++i) {
     for (int j=ystart; j<=yend; ++j) {
	dump_wgt_or_histo_weighted(i,j,thresh,NETWGT_EXC,fp);
     }
  }
  for (int i=xstart; i<=xend; ++i) {
     for (int j=ystart; j<=yend; ++j) {
	dump_wgt_or_histo_weighted(i,j,thresh,NETWGT_INH,fp);
     }
  }

  fclose(fp);

  return 1;

}

//----------------------------------------------------------------------------
//	plot_matrix(param_name) 
//	  - run ppmio and display ppm file using xv
//----------------------------------------------------------------------------
void netlayer::plot_matrix(char *param_name) {

  FILE *fp;
  char cmd[1024];
  char fname[1024];
  REAL imin,imax,rmin,rmax;

  int param_idx = net_bb->force_find(param_name);

  if (param_idx < 0) {
	cout << "netlayer::set_parameter() : no such parameter " << param_name
	     << endl;
 	return;
  }

  sprintf(fname,"/tmp/.netlayer.%s.mat",param_name);
  switch (net_bb->table[param_idx].type) {
      case BB_INT: break;
      case BB_IP: break;
      case BB_IPP: 
  		fp = fopen(fname,"w");
		matrix<int> *ibuffer;
		ibuffer = new matrix<int>(net_bb->table[param_idx].cols,
			net_bb->table[param_idx].rows);
		ibuffer->import_matrix(net_bb->table[param_idx].data.ipp);
		ibuffer->print_matrix(fp);
  		ibuffer->minmax(&imin,&imax);
  		cout << "* " << param_name << " " << ibuffer->sum() 
			<< " " << imin << " " << imax << endl;
  		fclose (fp);
		delete ibuffer;
		sprintf(cmd,"procppmmat %d %d %s",
			net_bb->table[param_idx].cols,
			net_bb->table[param_idx].rows,
			fname);
		system(cmd);
		break;
      case BB_REAL: break;
      case BB_RP: break;
      case BB_RPP: 
  		fp = fopen(fname,"w");
		matrix<REAL> *rbuffer;
		rbuffer = new matrix<REAL>(net_bb->table[param_idx].cols,
			net_bb->table[param_idx].rows);
		rbuffer->import_matrix(net_bb->table[param_idx].data.rpp);
		rbuffer->print_matrix(fp);
  		rbuffer->minmax(&rmin,&rmax);
  		cout << "* "<< param_name << " " << rbuffer->sum() 
			<< " " << rmin << " " << rmax << endl;
  		fclose (fp);
		delete rbuffer;
		sprintf(cmd,"procppmmat %d %d %s",
			net_bb->table[param_idx].cols,
			net_bb->table[param_idx].rows,
			fname);
		system(cmd);
		break;
  }
}
//----------------------------------------------------------------------------
int netlayer::call_plot_matrix(int argc, char *argv[]) {

  int cur_inp;

  if (argc!=1) 
	return 0;

  plot_matrix(argv[1]);

  return 1;

}
//----------------------------------------------------------------------------
int netlayer::call_plot_stat(int argc, char *argv[]) {

  char param_name[1024];
  int  col, row;

  if (argc!=2) 
	return 0;

  col = atoi(argv[1]);
  row = atoi(argv[2]);
  
  cerr << ">>> net_aff_input" << endl;
  plot_matrix("net_aff_input");

  cerr << ">>> net_exc_input" << endl;
  plot_matrix("net_exc_input");

  cerr << ">>> net_inh_input" << endl;
  plot_matrix("net_inh_input");

  cerr << ">>> wgts[" << col << "][" << row << "].aff_ws[0]" << endl;
  sprintf(param_name,"wgts[%d][%d].aff_ws[0]",col,row);
  plot_matrix(param_name);

  cerr << ">>> wgts[" << col << "][" << row << "].lat_exc_w" << endl;
  sprintf(param_name,"wgts[%d][%d].lat_exc_w",col,row);
  plot_matrix(param_name);

  cerr << ">>> wgts[" << col << "][" << row << "].lat_inh_w" << endl;
  sprintf(param_name,"wgts[%d][%d].lat_inh_w",col,row);
  plot_matrix(param_name);

  return 1;

}

//----------------------------------------------------------------------------
//	dump_wgt_or_stat() 
//	  - dump
//		1. phi : angle between the preferred orientation of
//			the source neuron and the line connecting
//			the two neurons (Direction)
//		2. theta : relative orientation preference difference
//			of the source and target neuron
//		3. d : distance between the two neurons
//
//	  - output format is 
//		
//		phi theta d
//
//		- all angles are in degrees.
//		phi : 0..360
//		theta : 0..180
//
//	  - The 3-tuple is printed out for weights exceeding thresh.
//
//----------------------------------------------------------------------------
void netlayer::dump_wgt_or_stat(int col, int row, REAL thresh,
		NETWGT_TYPE choice, FILE *fp, int omit_radius, 
		int angle_low, int angle_high) {

  // PRE-conditions: 
  //	Assume or_pref contains the current or_preference of this map
  //	Assume cog contains the current C.O.G of this map

  matrix<REAL> *buffer;
  // alloc buffer 
  buffer = new matrix<REAL>(cols,rows);

  int histo[35];

  switch (choice) {

  case NETWGT_EXC:
	  // 2. lat exc weights
	  // cout << "# excw : " << col << " " << row << endl;
	  // cout << cols << " " << rows << endl;
	  // 01feb21:  fprintf(fp,"# excw : %d %d\n",col,row);
	  get_lat_excw(buffer->mat,col,row,0);
	  break;
	  
  case NETWGT_INH:
	  // alloc buffer 
	  // 3. lat inh weights
	  // cout << "# inhw : " << col << " " << row << endl;
	  // cout << cols << " " << rows << endl;
	  fprintf(fp,"# inhw : %d %d\n",col,row);
	  get_lat_inhw(buffer->mat,col,row,0);
	  break;
  default: 
	  // not supported
	  cout << "netlayer::dump_wgt_or_stat : unsupported\n";
	  cerr << "netlayer::dump_wgt_or_stat : unsupported\n";
	  break;
  }

  // Now buffer contains the weight matrix.
  //
  // compare orientation of unit (col,row) and
  // other units connected with lateral connection
  // greater than "thresh" and increment count in
  // histogram.
  // The histogram is indexed by the difference in 
  // orientation. 

  int cur_pref = (int)(or_pref[col][row]*180.0+0.5);
  int phi,theta,d;
  REAL angle;
  REAL dest_col, dest_row;
  REAL src_col = cogs[0][col][row].x;
  REAL src_row = cogs[0][col][row].y;
  int  angle_offset;

  for (int i=0; i<cols; ++i) {
    for (int j=0; j<rows; ++j) {

	int src_angle = (cur_pref>90)?cur_pref-180:cur_pref;

	if (buffer->mat[i][j] > thresh 
		&& src_angle >= angle_low 
		&& src_angle <= angle_high) {

		// 1. Calculate relative orientation difference : theta
		int target_pref = (int)(or_pref[i][j]*180.0+0.5);
		if (target_pref<cur_pref) {
			theta = target_pref + 180 - cur_pref;
		} else {
			theta = target_pref - cur_pref;
		}

		// 2. Calculate distance : this must be the distance
		//	between afferent C.O.Gs.
		dest_col = cogs[0][i][j].x;
		dest_row = cogs[0][i][j].y;

		d = (int)(sqrt((dest_col-src_col)*(dest_col-src_col)
			  +(dest_row-src_row)*(dest_row-src_row))+0.5);

		// 3. Calculate phi, the direction of the target segment
		// relative to the current segment
		if (fabs(dest_col-src_col)<0.0001) {
			src_col += 0.0001;
			angle_offset = 180;
		} else {
			angle_offset = 0;
		}
		
		angle = (float)atan((double)(dest_row-src_row)/(dest_col-src_col)
			     )*180.0/M_PI + angle_offset;
		if (angle <0) angle += 360.0;

		phi = (int)(angle+0.5)-target_pref;

		if (phi<0) phi += 360;

		// 4. Print
	  	// fprintf(fp,"# cogs: %f %f %f %f %d\n",src_col,dest_col,src_row,dest_row,phi);
	  	fprintf(fp,"%d %d %d\n",phi,theta,d);
				
	}

    }

  }

  delete buffer;

}
//----------------------------------------------------------------------------
int netlayer::call_dump_wgt_or_stat(int argc, char *argv[]) {

  // call_dump_wgt_or_histo <startx> <starty> <endx> <endy> <thresh>

  calc_cog(0);


  if (argc!=5) {
	return 0;
  }

  int x1 = atoi(argv[1]); int y1 = atoi(argv[2]);
  int x2 = atoi(argv[3]); int y2 = atoi(argv[4]);
 
  REAL thresh = atof(argv[5]);

  int xstart = MIN(x1,x2); int ystart = MIN(y1,y2);
  int xend   = MAX(x1,x2); int yend   = MAX(y1,y2);

  FILE *fp = fopen("wgt_or.stat","w");

  if (!fp) {
	warning("call_dump_wgt_or_stat","cannot write to wgt_or.stat");
	return 0;
  }

  for (int i=xstart; i<=xend; ++i) {
     for (int j=ystart; j<=yend; ++j) {
	dump_wgt_or_stat(i,j,thresh,NETWGT_EXC,fp);
     }
  }
  /* 01feb21
  for (int i=xstart; i<=xend; ++i) {
     for (int j=ystart; j<=yend; ++j) {
	dump_wgt_or_stat(i,j,thresh,NETWGT_INH,fp);
     }
  }
  */

  fclose(fp);

  return 1;

}
//----------------------------------------------------------------------------
int netlayer::call_dump_wgt_or_stat_limited_angle(int argc, char *argv[]) {

  // call_dump_wgt_or_histo <startx> <starty> <endx> <endy> <thresh>
  // 			<low_angle> <high_angle>

  calc_cog(0);

  if (argc!=7) {
	return 0;
  }

  int x1 = atoi(argv[1]); int y1 = atoi(argv[2]);
  int x2 = atoi(argv[3]); int y2 = atoi(argv[4]);
 
  REAL thresh = atof(argv[5]);

  int low = atoi(argv[6]); int high = atoi(argv[7]);

  int xstart = MIN(x1,x2); int ystart = MIN(y1,y2);
  int xend   = MAX(x1,x2); int yend   = MAX(y1,y2);

  FILE *fp = fopen("wgt_or.stat","w");

  if (!fp) {
	warning("call_dump_wgt_or_stat","cannot write to wgt_or.stat");
	return 0;
  }

  for (int i=xstart; i<=xend; ++i) {
     for (int j=ystart; j<=yend; ++j) {
	dump_wgt_or_stat(i,j,thresh,NETWGT_EXC,fp,0,low,high);
     }
  }
  /* 01feb21
  for (int i=xstart; i<=xend; ++i) {
     for (int j=ystart; j<=yend; ++j) {
	dump_wgt_or_stat(i,j,thresh,NETWGT_INH,fp,low,high);
     }
  }
  */

  fclose(fp);

  return 1;

}


//----------------------------------------------------------------------------
//	netlayer::kill_wgt() 
//	  - kill 'type' connections for neuron (x,y) from 'cur_inp'th 
//	    input (cur_inp only for aff weights)
//	  - this kills outgoing weights for lateral connections,
//	    so incoming influence exists even after this
//----------------------------------------------------------------------------
void netlayer::kill_wgt(int x, int y, NETWGT_TYPE type, int cur_inp) {

  register REAL **wgt_pp;
  int  cols;
  int  rows;
  REAL wipeout_value;

  // 1. get pointer to the weight matrix and size.

  switch (type) {

	case NETWGT_AFF: 	
			wgt_pp = wgt[x][y].aff_ws[cur_inp]; 
		    	cols  = rf_ranges[cur_inp][x][y].cols;
			rows  = rf_ranges[cur_inp][x][y].rows;
			wipeout_value = 0.0;
			break;
	case NETWGT_INH:	
			wgt_pp = wgt[x][y].lat_inh_w; 
		    	cols  = lat_inh_range[x][y].cols;
			rows  = lat_inh_range[x][y].rows;
			wipeout_value = -DEATH_FLAG;
			break;
	case NETWGT_EXC:
			wgt_pp = wgt[x][y].lat_exc_w; 
		    	cols  = lat_exc_range[x][y].cols;
			rows  = lat_exc_range[x][y].rows;
			wipeout_value = -DEATH_FLAG;
			break;

	default: printf("kill_wgt: connection type %d not supported\n",type); 
			return; break;

  }

  // 2. wipe out 

  for (int i=0; i<cols; ++i) 
	for (int j=0; j<rows; ++j) 
		wgt_pp[i][j] = wipeout_value;

}
//----------------------------------------------------------------------------
void netlayer::kill_aff_wgt_range(int sx, int sy, int ex, int ey, 
  				     int cur_inp) {

   if (!bounds_check(sx,sy,ex,ey)) {
	printf("netlayer::disable_aff_wgt_range() - range out of bound.\n");
	fprintf(stderr,"netlayer::kill_aff_wgt_range() - range out of bound : %d %d -- %d %d\n",sx,sy,ex,ey); 
	return;
   }

   for (int i=sx; i<=ex; ++i)
	for (int j=sy; j<=ey; ++j) {
		kill_wgt(i,j,NETWGT_AFF,0);
	}

}
//----------------------------------------------------------------------------
// external interface for use with simenv class
int netlayer::call_kill_aff_wgt_range(int argc, char *argv[]) {

  if (argc!=5) return 0;

  kill_aff_wgt_range(atoi(argv[1]),atoi(argv[2]),atoi(argv[3]),
			atoi(argv[4]),atoi(argv[5]));

  return 1;

}

//----------------------------------------------------------------------------
void netlayer::kill_lat_exc_wgt_range(int sx, int sy, int ex, int ey) {

   if (!bounds_check(sx,sy,ex,ey)) {
	printf("netlayer::kill_lat_exc_wgt_range() - range out of bound.\n");
	fprintf(stderr,"netlayer::disable_aff_wgt_range() - range out of bound : %d %d -- %d %d\n",sx,sy,ex,ey); 
	return;
   }

   for (int i=sx; i<=ex; ++i)
	for (int j=sy; j<=ey; ++j) {
		kill_wgt(i,j,NETWGT_EXC);
	}

}
//----------------------------------------------------------------------------
// external interface for use with simenv class
int netlayer::call_kill_lat_exc_wgt_range(int argc, char *argv[]) {

  if (argc!=4) return 0;

  kill_lat_exc_wgt_range(atoi(argv[1]),atoi(argv[2]),atoi(argv[3]),
			atoi(argv[4]));

  return 1;

}


//----------------------------------------------------------------------------
//	kill_instar_lat_wgt() 
//	  - kill outgoing weights to a certain region
//----------------------------------------------------------------------------
void netlayer::kill_instar_lat_wgt(int i, int j, NETWGT_TYPE type){

  int   basex, basey;
  int   lowx,highx,lowy,highy;
  REAL  tmp_wgt;
  REAL  tmp_out;
  REAL  **tmp_wgt_pp;

   // 1.1 For all excitatory lateral connections
   //   later than: v1.12 use instar calculation here.

   lowx         = lat_exc_range[i][j].lowx;
   highx        = lat_exc_range[i][j].highx;
   lowy         = lat_exc_range[i][j].lowy;
   highy        = lat_exc_range[i][j].highy;

   switch (type) {

	case NETWGT_EXC:
   		for (int x=lowx; x<=highx; ++x) {
      			for (int y=lowy; y<=highy; ++y) {
				int lx = lat_exc_range[x][y].lowx;
				int ly = lat_exc_range[y][y].lowy;
				wgt[x][y].lat_exc_w[i-lx][j-ly]=-DEATH_FLAG;
			}
		}
		break;

	case NETWGT_INH:
   		for (int x=lowx; x<=highx; ++x) {
      			for (int y=lowy; y<=highy; ++y) {
				int lx = lat_inh_range[x][y].lowx;
				int ly = lat_inh_range[y][y].lowy;
				wgt[x][y].lat_inh_w[i-lx][j-ly]=-DEATH_FLAG;
			}
		}
		break;

	default: printf("kill_instar_lat_wgt: unsupported type\n");
		break;
   }

}
//----------------------------------------------------------------------------
void netlayer::kill_instar_lat_exc_wgt_range(int sx, int sy, int ex, int ey) {

   if (!bounds_check(sx,sy,ex,ey)) {
	printf("netlayer::kill_instar_lat_exc_wgt_range() - range out of bound.\n");
	fprintf(stderr,"netlayer::kill_instar_exc_wgt_range() - range out of bound : %d %d -- %d %d\n",sx,sy,ex,ey); 
	return;
   }

   for (int i=sx; i<=ex; ++i)
	for (int j=sy; j<=ey; ++j) {
		kill_instar_lat_wgt(i,j,NETWGT_EXC);
	}

}
//----------------------------------------------------------------------------
// external interface for use with simenv class
int netlayer::call_kill_instar_lat_exc_wgt_range(int argc, char *argv[]) {

  if (argc!=4) return 0;

  kill_instar_lat_exc_wgt_range(atoi(argv[1]),atoi(argv[2]),atoi(argv[3]),
			atoi(argv[4]));

  return 1;

}

//----------------------------------------------------------------------------
//	dump_matrix(param_name) 
//	  - run ppmio to write to file
//----------------------------------------------------------------------------
void netlayer::dump_matrix(char *param_name,int t) {

  FILE *fp;
  char cmd[1024];
  char fname[1024];
  REAL imin,imax,rmin,rmax;
  ppmio *io;

  int param_idx = net_bb->force_find(param_name);

  if (param_idx < 0) {
	cout << "netlayer::set_parameter() : no such parameter " << param_name
	     << endl;
 	return;
  }

  sprintf(fname,"%s.%05d.ppm",param_name,t);
  switch (net_bb->table[param_idx].type) {
      case BB_INT: break;
      case BB_IP: break;
      case BB_IPP: 
		io = new ppmio(net_bb->table[param_idx].data.ipp,
				net_bb->table[param_idx].cols,
				net_bb->table[param_idx].rows,256);
		io->ppmwrite(fname);
		delete io;
		break;
      case BB_REAL: break;
      case BB_RP: break;
      case BB_RPP: 
		io = new ppmio(net_bb->table[param_idx].data.rpp,
				net_bb->table[param_idx].cols,
				net_bb->table[param_idx].rows,256);
		io->ppmwrite(fname);
		delete io;
		break;
  }
}
//----------------------------------------------------------------------------
int netlayer::call_dump_matrix(int argc, char *argv[]) {

  int cur_inp;

  if (argc!=2) {
	cerr << "netlayer::call_dump_matrix : incorrect number of args" << endl;
	return 0;
  }

  dump_matrix(argv[1],atoi(argv[2]));

  return 1;

}
